/*****************************************
 Copyright  2001-2005
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/
/**
  @file   cache.c
  @brief  Simple cache for HTTP streaming

  @author Julien Lerouge
  @date   2005-03-02
*/

#include <stdio.h>
#include "cache.h"
#include "../include/rmlibhttp.h"

#if 0
#define CDB
#else
#undef CDB
#endif

#ifdef CDB
#define CACHEDEBUG ENABLE
#else
#define CACHEDEBUG DISABLE
#endif

#define MAXBUFFER 32

struct cache_s {
	RMuint8 *buffer;
	RMuint32 start[MAXBUFFER];
	RMuint32 time[MAXBUFFER];
	RMuint32 bufadr[MAXBUFFER];
	RMuint32 free_buffer;
	RMuint32 hit;
	RMuint32 miss;
	RMint32 last_removed;
	RMint32 nbuffer;
	RMuint32 buffersize;
};

struct cache_s * init_cache(RMint32 nbuffer, RMuint32 buffersize)
{
	struct cache_s * cache = NULL;
	RMint32 i;
	
	RMDBGLOG((ENABLE,"Open a HTTP cache of %ld bytes\n", nbuffer*buffersize));
	cache = RMMalloc(sizeof(struct cache_s));
	if (cache == NULL)
		goto ouch;

	cache->nbuffer = nbuffer;
	cache->buffersize = buffersize;

	cache->buffer = RMMalloc(buffersize*nbuffer);
	if (cache->buffer == NULL)
		goto ouch;

	for (i=0; i<nbuffer; i++){
		cache->start[i] = 0;
		cache->time[i] = 0;
		cache->bufadr[i] = (RMuint32)(cache->buffer + i*buffersize);
	}
	
	cache->free_buffer = 0xFFFFFFFF;
	cache->hit = 0;
	cache->miss = 0;
	cache->last_removed = -1;

	return(cache);
ouch:
	if (cache){
		if (cache->buffer)
			RMFree(cache->buffer);
		RMFree(cache);
	}
	return NULL;
}
			
void destroy_cache(struct cache_s *cache)
{
	if (cache == NULL)
		return;
	RMFree(cache->buffer);
	RMFree(cache);
}

static RMint32 get_free_buffer(struct cache_s * cache)
{
	RMuint32 free = 1;
	RMint32 i;
	
	for ( i=0 ; i<cache->nbuffer ; i++){
		if ((cache->free_buffer & free) == 0)
			free = free << 1;
		else
			return(i);
	}

	return(-1);
}

static inline RMint32 free_buffer_available(struct cache_s * cache)
{
	if (cache->nbuffer == 32)
		return cache->free_buffer;
	return (cache->free_buffer & ((1<<cache->nbuffer)-1));
}

static inline void mark_used(struct cache_s * cache, RMint32 index)
{
	cache->free_buffer &= (~(1 << index));
}
	
static inline void mark_free(struct cache_s * cache, RMint32 index)
{
	cache->free_buffer |= (1 << index);
}

static inline RMint32 is_used(struct cache_s * cache, RMint32 index)
{
	return((cache->free_buffer & (1<<index)) == 0);
}

static RMint32 iscached(struct cache_s * cache, RMuint32 start)
{
	RMint32 i;
	
	for (i = 0; i < cache->nbuffer; i++){
		if ( is_used(cache, i) &&
		    (start == cache->start[i]))
			return(i);
	}

	return(-1);
}

static RMint32 get_first(struct cache_s * cache)
{
	RMuint32 absolute_oldest = 0;
	RMint32 i, absolute_index = -1;

	for (i=0; i <cache->nbuffer; i++){
		if ((absolute_index == -1) || 
		    (cache->start[i] < absolute_oldest)){
			absolute_index = i;
			absolute_oldest = cache->start[i];
		}
	}

	return absolute_index;
}

static RMint32 get_lru(struct cache_s * cache)
{
	RMuint32 min = cache->miss;
	RMint32 index = -1;
	RMint32 i;

	for (i=0; i<cache->nbuffer; i++){
		if (cache->time[i] < min){
			min = cache->time[i];
			index = i;
		}
	}

	if (index < 0)
		index = get_first(cache);

	return index;
}
		
static void flush_cache(struct cache_s * cache)
{
	cache->free_buffer = 0xFFFFFFFF;
	fprintf(stderr,"HTTP cache flushed\n");
}

#ifdef CDB
static void print_buffer(struct cache_s * cache)
{
	RMint32 i;

	for (i=0;i<cache->nbuffer;i++)
		if (is_used(cache, i))
			fprintf(stderr,"  %p buffer %ld, %10lu->%10lu time=%lu\n",
			        cache, i, cache->start[i], 
				cache->start[i] + cache->buffersize - 1, cache->time[i]);
}
#endif

RMint32 get_buffer_cache(struct cache_s * cache, RMuint32 start, RMuint8 **buffer)
{
	RMint32 index = 0;
	RMint32 cur_perf;

	cur_perf = (cache->hit+cache->miss) == 0 ? 0 : cache->hit*100/(cache->hit+cache->miss);
	RMDBGLOG((CACHEDEBUG,"  %lu -> %lu (hit %lu, miss %lu, %ld percent cached) \n",
	          start, start + cache->buffersize - 1, cache->hit, cache->miss, cur_perf));
	index = iscached(cache, start);
	if (index < 0){
		if (free_buffer_available(cache)){
			/* Not cached, free buffer available */
			RMDBGLOG((CACHEDEBUG,"  not cached, free space\n"));
			index = get_free_buffer(cache);
			mark_used(cache, index);
			cache->last_removed = -1;
		}else{
			/* Not cached, no more buffer */
			RMDBGLOG((CACHEDEBUG,"  not cached, no free space\n"));
			index = get_lru(cache);
			if (index == cache->last_removed){
				/* Spinning on the same buffer, bad, flush */
				flush_cache(cache);
				mark_used(cache, index);
			}
			cache->last_removed = index;
				
		}
		cache->start[index] = start;
		cache->time[index] = cache->miss;
		*buffer = (void *)(cache->bufadr[index]);
		cache->miss++;
#ifdef CDB
		print_buffer(cache);
#endif
		return(-1);
	
	}

	/* Cached */
	cache->hit++;
	RMDBGLOG((CACHEDEBUG,"  cached\n"));
	*buffer = (void *)(cache->bufadr[index]);
	cache->time[index] = cache->miss;
	return(1);
}

