/*-
 * Copyright (c) 2000-2004 Dag-Erling Codan Smrgrav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>

/*
 * The following copyright applies to the base64 code:
 *
 *-
 * Copyright 1997 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that both the above copyright notice and this
 * permission notice appear in all copies, that both the above
 * copyright notice and this permission notice appear in all
 * supporting documentation, and that the name of M.I.T. not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  M.I.T. makes
 * no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * THIS SOFTWARE IS PROVIDED BY M.I.T. ``AS IS''.  M.I.T. DISCLAIMS
 * ALL EXPRESS OR IMPLIED WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
 * SHALL M.I.T. BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <locale.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "../include/rmlibhttp.h"
#include "cache.h"
#include "common.h"
#include "httperr.h"

/* Maximum number of redirects to follow */
#define MAX_REDIRECT 5

#define CACHE 

/* Symbolic names for reply codes we care about */
#define HTTP_OK			200
#define HTTP_PARTIAL		206
#define HTTP_MOVED_PERM		301
#define HTTP_MOVED_TEMP		302
#define HTTP_SEE_OTHER		303
#define HTTP_NEED_AUTH		401
#define HTTP_NEED_PROXY_AUTH	407
#define HTTP_BAD_RANGE		416
#define HTTP_PROTOCOL_ERROR	999

#define HTTP_REDIRECT(xyz) ((xyz) == HTTP_MOVED_PERM \
			    || (xyz) == HTTP_MOVED_TEMP \
			    || (xyz) == HTTP_SEE_OTHER)

#define HTTP_ERROR(xyz) ((xyz) > 400 && (xyz) < 599)

/*****************************************************************************
 * I/O functions for decoding chunked streams
 */

#define CONN(io)	((io)->request->conn)

struct httpio
{
	struct http_request_s 	*request;	/* HTTP request */
	RMint32			 eof;		/* end-of-file flag */
	RMint32			 error;		/* error flag */
	RMuint32		 chunksize;	/* remaining size of current chunk */
#ifdef HTTP_DEBUG_ENABLED
	RMuint32		 total;
#endif
};

static time_t my_timegm (struct tm *tm) {
	time_t ret;
	RMascii *tz;

	tz = getenv("TZ");
	setenv("TZ", "", 1);
	tzset();
	ret = mktime(tm);
	if (tz)
		setenv("TZ", tz, 1);
	else
		unsetenv("TZ");
	tzset();
	return ret;
}

struct http_request_s 
{
	struct url *url;
	const RMascii *op;
	struct url *purl;
	struct url_stat *us;
	RMHTTPFlags flags;
	RMint32 chunked;
	RMint32 direct;
	RMint32 noredirect;
	RMint32 verbose;
	RMint32 need_auth;
	RMint32 open;
	RMint32 need_reopen;
	RMint64 offset;
	RMint64 clength;
	RMint64 length;
	RMint64 size;			/* Total size of the requested file */
	time_t mtime;
	RMint32 e;
	struct url *new;
	conn_t *conn;
	RMascii *host;
	struct cache_s *cache;
	RMint32 last_read_short;
	RMascii *custom_header;		/* Custom header for this request */
	void *custom_cookie;        	/* Used by custom hooks */
	HttpHookOps *custom_hooks;  	/* Used by custom hooks */
};


static RMint32 __http_get_reply(struct http_request_s *request);
static RMint32 __http_get_headers(struct http_request_s *request);
static RMint32 __http_send_request(struct http_request_s *request);
static RMint32 __http_connect(struct http_request_s *request);
static RMint32 _http_single_request(struct http_request_s *request);
static RMint32 _http_seekn(void *cookie, RMint64 *position, RMint32 whence);
static RMint32 _http_closefn(void *v);
static RMint32 _http_readfn(void *v, RMuint8 *buf, RMuint32 len);

typedef struct {
	void *readfn;
	void *writefnL;
	void *seekfn;
	void *closefn;
} httpFileOps_s;

static httpFileOps_s p_httpFileOps = { (void *) _http_readfn, (void *) NULL, (void *) _http_seekn, (void *) _http_closefn };

void *httpFileOps = &p_httpFileOps;

/* Custom header */
static RMascii	*fetchCustomHeader = NULL;

/* Custom hooks */
static void *fetchCustomCookie = NULL;
static HttpHookOps *fetchCustomHooks = NULL;

/*
 * Get next chunk header
 */
static RMint32
_http_new_chunk(struct httpio *io)
{
	RMascii *p;

	if (_fetch_getln(CONN(io)) == -1)
		return (-1);

	if (CONN(io)->buflen < 2 || !isxdigit(*CONN(io)->buf))
		return (-1);

	for (p = (RMascii *)CONN(io)->buf; *p && !isspace(*p); ++p) {
		if (*p == ';')
			break;
		if (!isxdigit(*p))
			return (-1);
		if (isdigit(*p)) {
			io->chunksize = io->chunksize * 16 +
			    *p - '0';
		} else {
			io->chunksize = io->chunksize * 16 +
			    10 + tolower(*p) - 'a';
		}
	}

#ifdef HTTP_DEBUG_ENABLED
	io->total += io->chunksize;
	if (io->chunksize == 0)
		RMDBGLOG((HTTPDEBUG, "%s(): end of last chunk\n", __func__));
	else
		RMDBGLOG((HTTPDEBUG, "%s(): new chunk: %lu (%lu)\n",
			  __func__, (RMuint32)io->chunksize,
			  (RMuint32)io->total));
#endif

	return (io->chunksize);
}

/*
 * Fill the given buffer, do chunk decoding on the fly
 */
static RMint32
_http_fillbuf(struct httpio *io, RMuint32 len, RMuint8 *buffer)
{
	RMint32 read;
	
	if (io->error)
		return (-1);
	if (io->eof)
		return (0);

	if (io->request->chunked == 0) {
		read = _fetch_read(CONN(io), buffer, len);
		if (read  < 0){
			io->error = 1;
			return (-1);
		}
		return (read);
	}

	if (io->chunksize == 0) {
		switch (_http_new_chunk(io)) {
		case -1:
			io->error = 1;
			return (-1);
		case 0:
			io->eof = 1;
			return (0);
		}
	}

	if (len > io->chunksize)
		len = io->chunksize;
	read = _fetch_read(CONN(io), buffer, len);
	if (read == -1) {
		io->error = 1;
		return (-1);
	}
	io->chunksize -= read;

	if (io->chunksize == 0) {
		RMascii endl[2];

		if (_fetch_read(CONN(io), (RMuint8 *)endl, 2) != 2 ||
		    endl[0] != '\r' || endl[1] != '\n'){
			RMDBGLOG((ENABLE,"No valid chunk separator\n"));
			return (-1);
		}
	}

	return (read);
}

static RMint32 __http_read(void *v, RMuint8 *buf)
{
	struct httpio *io = (struct httpio *)v;
	struct http_request_s *request = io->request;
	RMint32 status;
	RMint32 len = request->url->offset_end - request->url->offset + 1;
	RMint32 nread = 0;
	RMint32 total_read = 0;
	RMint32 hook_read = 0;     /* Used by custom hooks */
	RMuint8 *hook_buf = NULL;  /* Used by custom hooks */

	if (len == 0)
		return 0;
	
	while(1) {
		if (request->need_reopen){
			RMDBGLOG((HTTPDEBUG,"Need to open a new connection\n"));
			status = _fetch_close(request->conn);
			if (status < 0)
				perror("Error closing previous connection, keep going");
			SAFE(__http_connect(request));
			/* New connection */
			request->need_reopen = 0;
			/* Reset the chunk size */
			io->chunksize = 0;
		}

		status = __http_send_request(request);
		if (status < 0){
			RMDBGLOG((HTTPDEBUG,"Error sending request : %ld, %ld, %ld\n", status, request->conn->err, fetchLastErrCode));
			if (fetchLastErrCode == FETCH_OK || fetchLastErrCode == FETCH_PIPE){
				/* EPIPE possible if connection is closed by peer */
				request->need_reopen = 1;
				continue;
			}else
				return(-1);
		}

		status = __http_get_reply(request);
		if (status < 0){
			RMDBGLOG((HTTPDEBUG,"Error getting reply : %ld, %ld, %ld\n", status, request->conn->err, fetchLastErrCode));
			if (fetchLastErrCode == FETCH_OK){
				request->need_reopen = 1;
				continue;
			}else
				return(-1);
		}

		status = __http_get_headers(request);
		if (status < 0){
			RMDBGLOG((HTTPDEBUG,"Error getting headers : %ld, %ld, %ld\n", status, request->conn->err, fetchLastErrCode));
			if (fetchLastErrCode == FETCH_OK){
				request->need_reopen = 1;
				continue;
			}else
				return(-1);
		}

		/* If custom decryption hooks are being used, ... */
		if ((request->custom_cookie != NULL) &&
		    (request->custom_hooks  != NULL) &&
		    (request->custom_hooks->preread  != NULL) &&
		    (request->custom_hooks->postread != NULL)) {
			RMint32 new_read_size;
			RMint32 more_data = 0;
			RMint32 result;

			/* Resync the decryption block boundary, if necessary */
			if (request->custom_hooks->reopen != NULL) {
				if ((result = request->custom_hooks->reopen(request->custom_cookie)) < 0)
					return result;
			}

			/* Truncate read size if necessary to prevent reading beyond a decryption block boundary */
			if ((new_read_size = request->custom_hooks->preread(request->custom_cookie, buf, len)) <= 0)
				return new_read_size;

			/* Read the start of the decryption block */
			if ((nread = _http_fillbuf(io, new_read_size, buf)) <= 0)
				return(-1);

			/* Because of HTTP chunking, a second read may be necessary */
			if (nread < new_read_size) {
				RMint32 delta_read = new_read_size - nread;

				if (_http_fillbuf(io, delta_read, buf + delta_read) != delta_read)
					return(-1);
			}

			/* Decrypt the data */
			if ((hook_read = request->custom_hooks->postread(request->custom_cookie,
									 buf,
									 new_read_size,
									 &more_data)) <= 0)
				return hook_read;

			/* Adjust the pointers */
			len -= hook_read;
			buf += hook_read;
			hook_buf = buf;
		}

		while (len > 0){
			nread = _http_fillbuf(io, len, buf);
			if (nread <= 0){
				RMDBGLOG((ENABLE,"nread <= 0 : %ld\n", total_read));
				/* For some servers (Intel DTCPIP for example),
				 * a bad range is not a problem, so the request
				 * above will not fail at EOF, but the read
				 * should at least ... */
				if (request->last_read_short){
					/* EOF */
					io->eof = 1;
					return total_read;
				}
				break;
			}
			total_read += nread;
			len -= nread;
			buf += nread;
		}

		/* If custom decryption hooks are being used, ... */
		if ((request->custom_cookie != NULL) &&
		    (request->custom_hooks  != NULL) &&
		    (request->custom_hooks->preread  != NULL) &&
		    (request->custom_hooks->postread != NULL)) {
			RMint32 more_data;
			RMint32 result;

			/* Decrypt the data */
			if ((result = request->custom_hooks->postread(request->custom_cookie,
								      hook_buf,
								      total_read,
								      &more_data)) <= 0)
				return result;

			/* The total read size includes the first read */
			total_read += hook_read;
		}

		if (!total_read && io->error){
			RMDBGLOG((ENABLE,"IO error %ld\n", io->error));
			return (-1);
		}

		if (nread == 0){
			/* Might be end of file or peer closed connection (time
			 * out, ...), so try to reopen, if it was end of file,
			 * it will fail */
			request->need_reopen = 1;
			request->last_read_short = 1;
			continue;
		}

		request->last_read_short = 0;
		return(total_read);
	}

	return(-1);
}

#define mmin(a,b) ((a)<(b))?(a):(b);


static RMint32 __http_read_cached(void *v, RMuint8 *buf, RMuint32 len)
{
	struct httpio *io = (struct httpio *)v;
	struct http_request_s *request = io->request;
	RMint32 cached, status;
	RMuint8 *buf_cache;
	RMint32 plen;
	RMint64 start_save, end_save;
	RMint64 start, offset;

	if ((request->url->offset > (request->size - 1)) || len == 0)
		return(0);

	/* Ensure request offset are valid */
	request->url->offset_end = request->url->offset + len - 1;
	if (request->url->offset_end > ( request->size - 1 )){
		request->url->offset_end = request->size - 1;
		len = request->url->offset_end - request->url->offset + 1;
	}
	/*fprintf(stderr,"Read %lu -> %lu (%ld)\n", 
	        (RMuint32) request->url->offset, 
		(RMuint32) request->url->offset_end,
		(RMuint32) request->url->offset_end - (RMuint32) request->url->offset + 1);*/
	
	/* Keep them for later */
	start_save = request->url->offset;
	end_save = request->url->offset_end;
	plen = end_save - start_save + 1;

	start = ( start_save & (~((1<<CACHELOG2)-1)));
	offset = start_save & ((1<<CACHELOG2)-1);

	do {
		RMint64 read = mmin(plen, CACHEBUFFERSIZE - offset);

		cached = get_buffer_cache(request->cache, start, &buf_cache);
		RMDBGLOG((HTTPDEBUG,"request = %p, cache = %p, buf_cache = %p\n", request, request->cache, buf_cache));
		if (cached < 0){
			/* Not cached */
			request->url->offset = start;
			request->url->offset_end = mmin(start+CACHEBUFFERSIZE-1, (request->size-1));
			status = __http_read(v, buf_cache);
			if (status < 0){
				RMDBGLOG((ENABLE,"Error reading %lu->%lu : %ld %ld %ld\n", 
						(RMuint32) request->url->offset, 
						(RMuint32) request->url->offset_end,
						status, request->conn->err, fetchLastErrCode));
				request->url->offset = start_save;
				request->url->offset_end = end_save;
				return status;
			}
		}
		bcopy(buf_cache+offset, buf, read);
		buf+=read;
		plen-=read;
		offset = 0;
		start += CACHEBUFFERSIZE;
	} while (plen > 0);

	request->url->offset = end_save + 1;
	return(end_save-start_save+1);
}


static RMint32
__http_read_uncached(void *v, RMuint8 *buf, RMuint32 len)
{
	struct httpio *io = (struct httpio *)v;
	struct http_request_s *request = io->request;
	RMint32 total_read = 0;
	RMint32 nread = 0;

	while (1){
		if (request->need_reopen){
			RMint32 status;
			
			RMDBGLOG((HTTPDEBUG,"Need to open a new connection\n"));
			status = _fetch_close(request->conn);
			if (status < 0)
				perror("Error closing previous connection, keep going");
			/* New connection */
			status = _http_single_request(request);
			/* No redirect at this point allowed */
			if (status <= 0){
				if (request->last_read_short){
					/* EOF */
					io->eof = 1;
					request->last_read_short = 0;
					return total_read;
				} else {
					/* Error */
					return(-1);
				}
			}
			request->need_reopen = 0;
			/* Reset the chunk size */
			io->chunksize = 0;
		}
	
		while (len > 0){
			nread = _http_fillbuf(io, len, buf);
			if (nread <= 0){
				RMDBGLOG((HTTPDEBUG,"nread <= 0 : %ld\n", total_read));
				/* For some servers (Intel DTCPIP for example),
				 * a bad range is not a problem, so the request
				 * above will not fail at EOF, but the read
				 * should at least ... */
				if (request->last_read_short)
					return total_read;
				break;
			}
			total_read += nread;
			len -= nread;
			buf += nread;
		}
		
		if (!total_read && io->error){
			RMDBGLOG((ENABLE,"IO error %ld\n", io->error));
			return (-1);
		}
		request->url->offset += total_read;

		if (nread == 0){
			/* Might be end of file or peer closed connection (time
			 * out, ...), so try to reopen, if it was end of file,
			 * it will fail */
			request->need_reopen = 1;
			request->last_read_short = 1;
			continue;
		}

		request->last_read_short = 0;
		return (total_read);
	}
	return(-1);
}

	
/*
 * Read function
 */
static RMint32
_http_readfn(void *v, RMuint8 *buf, RMuint32 len)
{
	struct httpio *io = (struct httpio *)v;
	struct http_request_s *request = io->request;

	if (io->error){
		RMDBGLOG((ENABLE,"IO error %ld\n", io->error));
		return (-1);
	}
	if (io->eof){
		RMDBGLOG((ENABLE,"IO EOF %ld\n", io->eof));
		return (0);
	}

	if (request->open)
		return (__http_read_cached(v, buf, len));

	/* If custom decryption hooks are being used, ... */
	if ((request->custom_cookie != NULL) &&
	    (request->custom_hooks  != NULL) &&
	    (request->custom_hooks->preread  != NULL) &&
	    (request->custom_hooks->postread != NULL)) {
		RMint32 result;
		RMint32 new_read_size;
		RMint32 nread;
		RMint32 actual_read;
		RMint32 more_data = 0;
		RMint32 bytes_left = len;
		RMint32 total_read = 0;
		RMuint8 *buf_ptr = buf;

		do {
			/* If reopening the connection and resynchronizing the decryption block boundary are necessary, ... */
			if ((request->need_reopen) && (request->custom_hooks->reopen != NULL)) {
				if ((result = request->custom_hooks->reopen(request->custom_cookie)) < 0)
					return result;
			}

			/* Truncate read size if necessary to prevent reading beyond a decryption block boundary */
			if ((new_read_size = request->custom_hooks->preread(request->custom_cookie, buf_ptr, bytes_left)) <= 0)
				return new_read_size;

			/* Fill up the buffer without going beyond the end of a decryption block boundary */
			if ((nread = __http_read_uncached(v, buf_ptr, new_read_size)) <= 0)
				return nread;

			/* Decrypt the data */
			if ((actual_read = request->custom_hooks->postread(request->custom_cookie, buf_ptr, nread, &more_data)) <= 0)
				return actual_read;

			/* Update the position */
			bytes_left -= actual_read;
			total_read += actual_read;
			buf_ptr    += actual_read;

		} while ((more_data != 0) && (bytes_left > 0));

		return(total_read);
	}

	return (__http_read_uncached(v, buf, len));
}

RMint32 fetchRead(RMuint8 *buf, RMint32 size, HTTPFile * file)
{
	return(_http_readfn((void *)file, buf, (RMuint32) size));
}

/*
 * Close function
 */
static RMint32
_http_closefn(void *v)
{
	struct httpio *io = (struct httpio *)v;
	RMint32 r;
	struct http_request_s *request = io->request;

	/* Perform a custom close hook if present */
	if ((request->custom_cookie != NULL) &&
	    (request->custom_hooks  != NULL) &&
	    (request->custom_hooks->close != NULL))
		request->custom_hooks->close(request->custom_cookie);

	r = _fetch_close(CONN(io));
	if (io->request->purl)
		fetchFreeURL(io->request->purl);
	if (io->request->url)
		fetchFreeURL(io->request->url);
	if (io->request->host)
		RMFree(io->request->host);
	if (io->request->cache)
		destroy_cache(io->request->cache);
	if (io->request->custom_header != NULL)
		RMFree(io->request->custom_header);
	if (io->request)
		RMFree(io->request);
	RMFree(io);
	return (r);
}

RMint32 fetchClose(HTTPFile * file)
{
	return (_http_closefn((void *) file));
}

/*
 * Seek function
 */
static RMint32 _http_seekn(void *cookie, RMint64 *position, RMint32 whence)
{
	struct httpio *io = (struct httpio *)cookie;
	struct http_request_s *request = io->request;
	RMint64 newpos = -1;

	/* Perform a custom seek hook if present */
	if ((request->custom_cookie != NULL) &&
	    (request->custom_hooks  != NULL) &&
	    (request->custom_hooks->seek != NULL))
		request->custom_hooks->seek(request->custom_cookie, *position, whence);

	if (whence == SEEK_SET){
		newpos = *position;
	}else if (whence == SEEK_CUR){
		newpos = request->url->offset + *position;
	}else if (whence == SEEK_END){
		newpos = (RMint64) request->size + *position;
	}

	RMDBGLOG((HTTPDEBUG,"Seeking : newpos = %lld pos = %lld\n", newpos, *position));

	if (newpos < 0){
		RMDBGLOG((ENABLE,"Seeking out of bounds : %lld\n", newpos));
		return(-1);
	}
	
	/* See man lseek, seeking beyond end of file should not return an error
	 * (allow creation of sparse file for example) */
#if 0
	if ( (request->size != -1) && (newpos > request->size) ){
		RMDBGLOG((ENABLE,"Seeking out of bounds : size = %lld, newpos = %lld\n", request->size, newpos));
		return(-1);
	}
#endif

	if ((request->url->offset != newpos) && 
	    (request->open == 0)){
		/* Closed connection request, we need to reopen a connection */
		if (request->open == 0)
			request->need_reopen = 1;
	}
	request->url->offset = newpos;
	*position = newpos;
	io->eof = 0;
	return(0);
}


RMint32 fetchSeek(HTTPFile * file, RMint64 offset, RMint32 whence)
{
	RMint64 pos=offset;
	return(_http_seekn((void *) file, &pos, whence));
}

RMint64 fetchTell(HTTPFile * file){
	RMint64 pos=0;
	RMint32 status;
	
	SAFE(_http_seekn((void *) file, &pos, SEEK_CUR));

	return pos;
}
			
/*
 * Wrap it up
 */
static HTTPFile *
_http_funopen(struct http_request_s *request)
{
	struct httpio *io;

	if (request == NULL)
		return(NULL);

	if ((io = RMCalloc(1, sizeof(*io))) == NULL) {
		_fetch_syserr();
		return (NULL);
	}
	io->request = request;

	return (HTTPFile *)io;
}


/*****************************************************************************
 * Helper functions for talking to the server and parsing its replies
 */

/* Header types */
typedef enum {
	hdr_syserror = -2,
	hdr_error = -1,
	hdr_end = 0,
	hdr_unknown = 1,
	hdr_content_length,
	hdr_content_range,
	hdr_last_modified,
	hdr_location,
	hdr_transfer_encoding,
	hdr_www_authenticate,
	hdr_connection
} hdr_t;

/* Names of interesting headers */
static struct {
	hdr_t		 num;
	const RMascii	*name;
} hdr_names[] = {
	{ hdr_content_length,		"Content-Length" },
	{ hdr_content_range,		"Content-Range" },
	{ hdr_last_modified,		"Last-Modified" },
	{ hdr_location,			"Location" },
	{ hdr_transfer_encoding,	"Transfer-Encoding" },
	{ hdr_www_authenticate,		"WWW-Authenticate" },
	{ hdr_connection,		"Connection"},
	{ hdr_unknown,			NULL },
};

/*
 * Send a formatted line; optionally echo to terminal
 */
static RMint32
_http_cmd(conn_t *conn, const RMascii *fmt, ...)
{
#define MAXLINE 1024
	va_list ap;
	RMuint32 len;
	RMascii msg[MAXLINE];
	RMascii *long_msg = NULL;
	RMint32 r;
	RMuint32 long_size = MAXLINE;

	va_start(ap, fmt);
	len = vsnprintf(msg, MAXLINE, fmt, ap);
	va_end(ap);

	if (len >= MAXLINE){
		do {
			if (long_msg)
				RMFree(long_msg);
			long_size = long_size * 2;
			long_msg = RMMalloc(long_size);
			if (long_msg == NULL){
				errno = ENOMEM;
				goto error;
			}
			va_start(ap, fmt);
			len = vsnprintf(long_msg, long_size, fmt, ap);
			va_end(ap);
		} while (len >= long_size);
	}

	if (long_msg)
		r = _fetch_putln(conn, long_msg, len);
	else
		r = _fetch_putln(conn, msg, len);
	if (r < 0) 
		goto error;
	
	return (0);

error:
	_fetch_syserr();
	if (long_msg)
		RMFree(long_msg);
	return(-1);
	
}

/*
 * Get and parse status line
 */
static RMint32
_http_get_reply(conn_t *conn)
{
	RMascii *p;

	if (_fetch_getln(conn) == -1){
		return (-1);
	}
	/*
	 * A valid status line looks like "HTTP/m.n xyz reason" where m
	 * and n are the major and minor protocol version numbers and xyz
	 * is the reply code.
	 * Unfortunately, there are servers out there (NCSA 1.5.1, to name
	 * just one) that do not send a version number, so we can't rely
	 * on finding one, but if we do, insist on it being 1.0 or 1.1.
	 * We don't care about the reason phrase.
	 */
	if (strncmp((RMascii *)conn->buf, "HTTP", 4) != 0)
		return (HTTP_PROTOCOL_ERROR);
	p = (RMascii *)conn->buf + 4;
	if (*p == '/') {
		if (p[1] != '1' || p[2] != '.' || (p[3] != '0' && p[3] != '1'))
			return (HTTP_PROTOCOL_ERROR);
		p += 4;
	}
	if (*p != ' ' || !isdigit(p[1]) || !isdigit(p[2]) || !isdigit(p[3]))
		return (HTTP_PROTOCOL_ERROR);

	conn->err = (p[1] - '0') * 100 + (p[2] - '0') * 10 + (p[3] - '0');
	return (conn->err);
}

/*
 * Check a header; if the type matches the given string, return a pointer
 * to the beginning of the value.
 */
static const RMascii *
_http_match(const RMascii *str, const RMascii *hdr)
{
	while (*str && *hdr && tolower(*str++) == tolower(*hdr++))
		/* nothing */;
	if (*str || *hdr != ':')
		return (NULL);
	while (*hdr && isspace(*++hdr))
		/* nothing */;
	return (hdr);
}

/*
 * Get the next header and return the appropriate symbolic code.
 */
static hdr_t
_http_next_header(conn_t *conn, const RMascii **p)
{
	RMint32 i;

	if (_fetch_getln(conn) == -1)
		return (hdr_syserror);
	while (conn->buflen && isspace(conn->buf[conn->buflen - 1]))
		conn->buflen--;
	conn->buf[conn->buflen] = '\0';
	if (conn->buflen == 0)
		return (hdr_end);
	/*
	 * We could check for malformed headers but we don't really care.
	 * A valid header starts with a token immediately followed by a
	 * colon; a token is any sequence of non-control, non-whitespace
	 * characters except "()<>@,;:\\\"{}".
	 */
	for (i = 0; hdr_names[i].num != hdr_unknown; i++)
		if ((*p = _http_match(hdr_names[i].name, (RMascii *)conn->buf)) != NULL)
			return (hdr_names[i].num);
	return (hdr_unknown);
}

/*
 * Parse a last-modified header
 */
static RMint32
_http_parse_mtime(const RMascii *p, time_t *mtime)
{
	RMascii locale[64], *r;
	struct tm tm;

	strncpy(locale, setlocale(LC_TIME, NULL), sizeof(locale));
	setlocale(LC_TIME, "C");
	r = (RMascii *) strptime(p, "%a, %d %b %Y %H:%M:%S GMT", &tm);
	/* XXX should add support for date-2 and date-3 */
	setlocale(LC_TIME, locale);
	if (r == NULL)
		return (-1);
	RMDBGLOG((HTTPDEBUG,"    last modified: [%04d-%02d-%02d "
		  "%02d:%02d:%02d]\n",
		  tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
		  tm.tm_hour, tm.tm_min, tm.tm_sec));
	*mtime = my_timegm(&tm);
	return (0);
}

/*
 * Parse a content-length header
 */
static RMint32
_http_parse_length(const RMascii *p, RMint64 *length)
{
	RMint64 len;

	for (len = 0; *p && isdigit(*p); ++p)
		len = len * 10 + (*p - '0');
	if (*p)
		return (-1);
	RMDBGLOG((HTTPDEBUG, "    content length: [%lld]\n",
	    (RMint64)len));
	*length = len;
	return (0);
}

/*
 * Parse a content-range header
 */
static RMint32
_http_parse_range(const RMascii *p, RMint64 *offset, RMint64 *length, RMint64 *size)
{
	RMint64 first, last, len;

	if (strncasecmp(p, "bytes ", 6) != 0)
		return (-1);
	p += 6;
	if (*p == '*') {
		first = last = -1;
		++p;
	} else {
		for (first = 0; *p && isdigit(*p); ++p)
			first = first * 10 + *p - '0';
		if (*p != '-')
			return (-1);
		for (last = 0, ++p; *p && isdigit(*p); ++p)
			last = last * 10 + *p - '0';
	}
	if (first > last || *p != '/')
		return (-1);
	for (len = 0, ++p; *p && isdigit(*p); ++p)
		len = len * 10 + *p - '0';
	if (*p || len < last - first + 1)
		return (-1);
	if (first == -1) {
		RMDBGLOG((HTTPDEBUG, "    content range: [*/%lld]\n",
		    (RMint64)len));
		*length = 0;
	} else {
		RMDBGLOG((HTTPDEBUG, "    content range: [%lld-%lld/%lld]\n",
		    (RMint64)first, (RMint64)last, (RMint64)len));
		*length = last - first + 1;
	}
	*offset = first;
	*size = len;
	return (0);
}


/*****************************************************************************
 * Helper functions for authorization
 */

/*
 * Base64 encoding
 */
static RMascii *
_http_base64(const RMascii *src)
{
	static const RMascii base64[] =
	    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	    "abcdefghijklmnopqrstuvwxyz"
	    "0123456789+/";
	RMascii *str, *dst;
	RMuint32 l;
	RMint32 t, r;

	l = strlen(src);
	if ((str = RMMalloc(((l + 2) / 3) * 4 + 1)) == NULL)
		return (NULL);
	dst = str;
	r = 0;

	while (l >= 3) {
		t = (src[0] << 16) | (src[1] << 8) | src[2];
		dst[0] = base64[(t >> 18) & 0x3f];
		dst[1] = base64[(t >> 12) & 0x3f];
		dst[2] = base64[(t >> 6) & 0x3f];
		dst[3] = base64[(t >> 0) & 0x3f];
		src += 3; l -= 3;
		dst += 4; r += 4;
	}

	switch (l) {
	case 2:
		t = (src[0] << 16) | (src[1] << 8);
		dst[0] = base64[(t >> 18) & 0x3f];
		dst[1] = base64[(t >> 12) & 0x3f];
		dst[2] = base64[(t >> 6) & 0x3f];
		dst[3] = '=';
		dst += 4;
		r += 4;
		break;
	case 1:
		t = src[0] << 16;
		dst[0] = base64[(t >> 18) & 0x3f];
		dst[1] = base64[(t >> 12) & 0x3f];
		dst[2] = dst[3] = '=';
		dst += 4;
		r += 4;
		break;
	case 0:
		break;
	}

	*dst = 0;
	return (str);
}

/*
 * Encode username and password
 */
static RMint32
_http_basic_auth(conn_t *conn, const RMascii *hdr, const RMascii *usr, const RMascii *pwd)
{
	RMascii *upw, *auth;
	RMint32 r;

	RMDBGLOG((HTTPDEBUG, "usr: [%s]\n", usr));
	RMDBGLOG((HTTPDEBUG, "pwd: [%s]\n", pwd));
	upw = RMMalloc(strlen(usr)+strlen(pwd)+2);
	if (sprintf(upw, "%s:%s", usr, pwd) == -1)
		return (-1);
	auth = _http_base64(upw);
	RMFree(upw);
	if (auth == NULL)
		return (-1);
	r = _http_cmd(conn, "%s: Basic %s", hdr, auth);
	RMFree(auth);
	return (r);
}

/*
 * Send an authorization header
 */
static RMint32
_http_authorize(conn_t *conn, const RMascii *hdr, const RMascii *p)
{
	/* basic authorization */
	if (strncasecmp(p, "basic:", 6) == 0) {
		RMascii *user, *pwd, *str;
		RMint32 r;

		/* skip realm */
		for (p += 6; *p && *p != ':'; ++p)
			/* nothing */ ;
		if (!*p || strchr(++p, ':') == NULL)
			return (-1);
		str = RMMallocAndDuplicateAscii(p);
		if (str == NULL)
			return (-1); /* XXX */
		user = str;
		pwd = strchr(str, ':');
		*pwd++ = '\0';
		r = _http_basic_auth(conn, hdr, user, pwd);
		RMFree(str);
		return (r);
	}
	return (-1);
}


/*****************************************************************************
 * Helper functions for connecting to a server or proxy
 */

/*
 * Connect to the correct HTTP server or proxy.
 */
static conn_t *
_http_connect(struct url *URL, struct url *purl, RMHTTPFlags flags)
{
	conn_t *conn;
	RMint32 verbose;
	RMint32 af;

#ifdef INET6
	af = AF_UNSPEC;
#else
	af = AF_INET;
#endif

	verbose = (flags & RM_HTTP_VERBOSE);
	if (flags & RM_HTTP_IPV4)
		af = AF_INET;
#ifdef INET6
	else if (flags & RM_HTTP_IPV6)
		af = AF_INET6;
#endif

	if (purl && strcasecmp(URL->scheme, SCHEME_HTTPS) != 0) {
		URL = purl;
	} else if (strcasecmp(URL->scheme, SCHEME_FTP) == 0) {
		/* can't talk http to an ftp server */
		/* XXX should set an error code */
		return (NULL);
	}

	if ((conn = _fetch_connect(URL->host, URL->port, af, verbose)) == NULL)
		/* _fetch_connect() has already set an error code */
		return (NULL);
	if (strcasecmp(URL->scheme, SCHEME_HTTPS) == 0 &&
	    _fetch_ssl(conn, verbose) == -1) {
		_fetch_close(conn);
		/* grrr */
		errno = ECONNREFUSED;
		_fetch_syserr();
		return (NULL);
	}
	return (conn);
}

static struct url *
_http_get_proxy(RMHTTPFlags flags)
{
	struct url *purl;
	RMascii *p;

	if (flags & RM_HTTP_NO_PROXY)
		return (NULL);
	if (((p = getenv("HTTP_PROXY")) || (p = getenv("http_proxy"))) &&
	    (purl = fetchParseURL(p))) {
		if (!*purl->scheme)
			strcpy(purl->scheme, SCHEME_HTTP);
		if (!purl->port)
			purl->port = _fetch_default_proxy_port(purl->scheme);
		if (strcasecmp(purl->scheme, SCHEME_HTTP) == 0)
			return (purl);
		fetchFreeURL(purl);
	}
	return (NULL);
}

/*****************************************************************************
 * Core
 */

/**
 * Connect to a specified server
 * @param request - the HTTP request
 * @return < 0 on error
 */
static RMint32 __http_connect(struct http_request_s *request)
{
	struct url *url, *purl;

	if (request == NULL)
		return(-1);

	url = request->url;
	purl = request->purl;

	/* check port */
	if (!url->port)
		url->port = _fetch_default_port(url->scheme);

	/* were we redirected to an FTP URL? */
	if (purl == NULL && strcmp(url->scheme, SCHEME_FTP) == 0) {
		RMDBGLOG((ENABLE,"Cannot redirect HTTP to FTP\n"));
		return(-1);
	}

	/* connect to server or proxy */
	if ((request->conn = _http_connect(url, purl, request->flags)) == NULL)
		return(-1);
	
	if (request->host)
		RMFree(request->host);
	request->host = (RMascii *) RMMalloc(strlen(url->host)+16);
	strcpy(request->host, url->host);
#ifdef INET6
	if (strchr(url->host, ':')) {
		snprintf(request->host, strlen(url->host)+15, "[%s]", url->host);
	}
#endif
	if (url->port != _fetch_default_port(url->scheme)) {
		snprintf(request->host + strlen(request->host), 15, ":%ld", url->port);
	}

	return(0);
}

/** 
 * Send a HTTP request
 * @param request - the HTTP request
 * @return <0 on error
 */
static RMint32 __http_send_request(struct http_request_s *request)
{
	struct url *url, *purl;
	conn_t *conn;
	const RMascii *p;
	RMint32 status;

	if (request == NULL)
		return(-1);

	url = request->url;
	purl = request->purl;
	conn = request->conn;

	/* send request */
	if (request->verbose)
		_fetch_info("requesting %s://%s%s",
		    url->scheme, request->host, url->doc);
	if (purl) {
		SAFE(_http_cmd(conn, "%s %s://%s%s HTTP/1.1",
		    request->op, url->scheme, request->host, url->doc));
	} else {
		SAFE(_http_cmd(conn, "%s %s HTTP/1.1",
		    request->op, url->doc));
	}

	/* virtual host */
	SAFE(_http_cmd(conn, "Host: %s", request->host));

	/* proxy authorization */
	if (purl) {
		if (*purl->user || *purl->pwd){
			SAFE(_http_basic_auth(conn, "Proxy-Authorization",
			    purl->user, purl->pwd));
		} else if ((p = getenv("HTTP_PROXY_AUTH")) != NULL && *p != '\0') {
			SAFE(_http_authorize(conn, "Proxy-Authorization", p));
		}
	}

	/* server authorization */
	if (request->need_auth || *url->user || *url->pwd) {
		if (*url->user || *url->pwd){
			SAFE(_http_basic_auth(conn, "Authorization", url->user, url->pwd));
		} else if ((p = getenv("HTTP_AUTH")) != NULL && *p != '\0') {
			SAFE(_http_authorize(conn, "Authorization", p));
		} else {
			_http_seterr(HTTP_NEED_AUTH);
			return(-1);
		}
	}

	/* other headers */
	if ((p = getenv("HTTP_REFERER")) != NULL && *p != '\0') {
		if (strcasecmp(p, "auto") == 0) {
			SAFE(_http_cmd(conn, "Referer: %s://%s%s", url->scheme, request->host, url->doc));
		} else {
			SAFE(_http_cmd(conn, "Referer: %s", p));
		}
	}

	/* Custom header */
	if (request->custom_header){
		SAFE(_http_cmd(conn,"%s", request->custom_header));
	}
	
	if ((p = getenv("HTTP_USER_AGENT")) != NULL && *p != '\0')
		SAFE(_http_cmd(conn, "User-Agent: %s", p));
	/*else
		SAFE(_http_cmd(conn, "User-Agent: %s " _LIBFETCH_VER, getprogname()));*/

	if (url->offset > 0 || request->open){
		if (request->open){
			SAFE(_http_cmd(conn, "Range: bytes=%lld-%lld", (RMint64)url->offset, (RMint64)url->offset_end));
		} else {
			SAFE(_http_cmd(conn, "Range: bytes=%lld-", (RMint64)url->offset));
		}
	}

	if (!request->open){
		SAFE(_http_cmd(conn, "Connection: close"));
	}
	SAFE(_http_cmd(conn, ""));

	return(0);
}

/**
 * Get HTTP reply from the server
 * @param request - the HTTP request
 * @return < 0 on error
 */
static RMint32 __http_get_reply(struct http_request_s *request)
{
	conn_t *conn;

	if (request == NULL){
		/* RMDBGLOG((ENABLE,"Null request\n")); */
		return(-1);
	}

	conn = request->conn;
	
	/* get reply */
	switch (_http_get_reply(conn)) {
	case HTTP_OK:
		if (request->url && request->url->offset > 0){
			/* We should have received a HTTP_PARTIAL, see RFC
			 * 2616, section 14.35 */
			RMDBGLOG((HTTPDEBUG,"HTTP_OK received instead of HTTP_PARTIAL, range not satisfied\n"));
			request->last_read_short = 1;
			return(-1);
		}
		break;
	case HTTP_PARTIAL:
		/* fine */
		break;
	case HTTP_MOVED_PERM:
	case HTTP_MOVED_TEMP:
	case HTTP_SEE_OTHER:
		/*
		 * Not so fine, but we still have to read the
		 * headers to get the new location.
		 */
		break;
	case HTTP_NEED_AUTH:
		if (request->need_auth) {
			/*
			 * We already sent out authorization code,
			 * so there's nothing more we can do.
			 */
			_http_seterr(conn->err);
			return(-1);
		}
		/* try again, but send the password this time */
		if (request->verbose)
			_fetch_info("server requires authorization");
		break;
	case HTTP_NEED_PROXY_AUTH:
		/*
		 * If we're talking to a proxy, we already sent
		 * our proxy authorization code, so there's
		 * nothing more we can do.
		 */
		_http_seterr(conn->err);
		return(-1);
	case HTTP_BAD_RANGE:
		/*
		 * This can happen if we ask for 0 bytes because
		 * we already have the whole file.  Consider this
		 * a success for now, and check sizes later.
		 */
		request->last_read_short = 1;
		break;
	case HTTP_PROTOCOL_ERROR:
		/* fall through */
	case -1:
		/* RMDBGLOG((ENABLE,"Error getting reply : errno = %d\n", errno)); */
		_fetch_syserr();
		return(-1);
	default:
		_http_seterr(conn->err);
		if (!request->verbose)
			return (-1);
		/* fall through so we can get the full error message */
	}
	return(0);
}

/**
 * Get headers from a HTTP reply
 * @param request - the HTTP request
 * @return < 0 on error
 */
static RMint32 __http_get_headers(struct http_request_s *request)
{
	struct url *url, *purl;
	conn_t *conn;
	const RMascii *p;
	hdr_t h;

	if (request == NULL)
		return(-1);

	url = request->url;
	purl = request->purl;
	conn = request->conn;

	/* get headers */
	do {
		switch ((h = _http_next_header(conn, &p))) {
		case hdr_syserror:
			_fetch_syserr();
			return (-1);
		case hdr_error:
			_http_seterr(HTTP_PROTOCOL_ERROR);
			return (-1);
		case hdr_content_length:
			_http_parse_length(p, &request->clength);
			break;
		case hdr_content_range:
			_http_parse_range(p, &request->offset, &request->length, &request->size);
			break;
		case hdr_last_modified:
			_http_parse_mtime(p, &request->mtime);
			break;
		case hdr_location:
			if (!HTTP_REDIRECT(conn->err))
				break;
			if (request->new)
				RMFree(request->new);
			if (request->verbose)
				_fetch_info("%d redirect to %s", conn->err, p);
			if (*p == '/')
				/* absolute path */
				request->new = fetchMakeURL(url->scheme, url->host, url->port, p,
				    url->user, url->pwd);
			else
				request->new = fetchParseURL(p);
			if (request->new == NULL) {
				/* XXX should set an error code */
				RMDBGLOG((ENABLE, "failed to parse new URL\n"));
				return(-1);
			}
			if (!*request->new->user && !*request->new->pwd) {
				strcpy(request->new->user, url->user);
				strcpy(request->new->pwd, url->pwd);
			}
			request->new->offset = url->offset;
			request->new->length = url->length;
			break;
		case hdr_transfer_encoding:
			/* XXX weak test*/
			request->chunked = (strcasecmp(p, "chunked") == 0);
			/* RMDBGLOG((HTTPDEBUG,"Chunked !!! %d\n", request->chunked)); */
			break;
		case hdr_www_authenticate:
			if (conn->err != HTTP_NEED_AUTH)
				break;
			/* if we were smarter, we'd check the method and realm */
			break;
		case hdr_connection:
			/* Assume the connection will close, need to reopen in
			 * case of open connection */
			if (request->open)
				request->need_reopen = 1;
		case hdr_end:
			/* fall through */
		case hdr_unknown:
			/* ignore */
			break;
		}
	} while (h > hdr_end);

	return(0);
}


	
/**
 * Send a single HTTP request
 * @param request - the HTTP request
 * @return -1 on error, 0 if redirect, 1 if HIT. In case of a redirect, the
 * "new" field is set to the redirected URL.
 */ 
static RMint32 _http_single_request(struct http_request_s *request)
{
	RMint32 status;
 
	if (request == NULL)
		return(-1);

	request->new = NULL;
	request->chunked = 0;
	request->offset = 0;
	request->clength = -1;
	request->length = -1;
	request->mtime = 0;
	
	status = __http_connect(request);
	if (status < 0)
		return(-1);
	
	status = __http_send_request(request);
	if (status < 0)
		return(status);

	status = __http_get_reply(request);
	if (status < 0)
		return(status);

	status = __http_get_headers(request);
	if (status < 0)
		return(status);

	/* we need to provide authentication */
	if (request->conn->err == HTTP_NEED_AUTH) {
		request->e = request->conn->err;
		request->need_auth = 1;
		_fetch_close(request->conn);
		request->conn = NULL;
		return(0);
	}

	/* requested range not satisfiable */
	if (request->conn->err == HTTP_BAD_RANGE) {
		if (request->url->offset == request->size && request->url->length == 0) {
			/* asked for 0 bytes; fake it */
			request->offset = request->url->offset;
			request->conn->err = HTTP_OK;
			return(1);
		} else {
			_http_seterr(request->conn->err);
			return(-1);
		}
	}

	/* we have a hit or an error */
	if (request->conn->err == HTTP_OK || 
	    request->conn->err == HTTP_PARTIAL || 
	    HTTP_ERROR(request->conn->err))
		return(1);

	/* all other cases: we got a redirect */
	request->e = request->conn->err;
	request->need_auth = 0;
	_fetch_close(request->conn);
	request->conn = NULL;
	if (!request->new) {
		RMDBGLOG((ENABLE, "redirect with no new location\n"));
		return(1);
	}

	return(0);
}

/*
 * Send a request and process the reply
 *
 * @param URL - url to fetch
 * @param op - operation requested (GET, POST, ...)
 * @param us - url stat
 * @param purl - proxy url
 * @param flasg
 * @return HTTPFile *
 *
 */
static HTTPFile *
_http_request(struct url *URL, const RMascii *op, struct url_stat *us,
    struct url *purl, RMHTTPFlags flags)
{
	struct http_request_s *request;
	RMint32 i, n, status;
	HTTPFile *f;

	request = (struct http_request_s *) RMMalloc(sizeof(struct http_request_s));
	if (request == NULL){
		RMDBGLOG((ENABLE,"Cannot malloc\n"));
		return NULL;
	}

	RMMemset(request, 0, sizeof(struct http_request_s));

	/* Handle the flags */
	request->direct = (flags & RM_HTTP_NO_PROXY);
	request->noredirect = (flags & RM_HTTP_NO_REDIRECT);
	request->verbose = (flags & RM_HTTP_VERBOSE);
	request->open = (flags & RM_HTTP_OPEN_CACHED);
	if (request->open){
		request->cache = init_cache(CACHENBUFFER, CACHEBUFFERSIZE);
		if (request->cache == NULL)
			goto ouch;
	}
	if ((flags & RM_HTTP_CUSTOM_HEADER) && (fetchCustomHeader != NULL)){
		request->custom_header = RMMallocAndDuplicateAscii(fetchCustomHeader);
	}
	if ((flags & RM_HTTP_CUSTOM_HOOKS) && (fetchCustomCookie != NULL) && (fetchCustomHooks != NULL)){
		request->custom_cookie = fetchCustomCookie;
		request->custom_hooks = fetchCustomHooks;
		fetchCustomCookie = NULL;
		fetchCustomHooks = NULL;

		/* Perform a custom open hook if present */
		if (request->custom_hooks->open != NULL)
			request->custom_hooks->open(request->custom_cookie);
	}
	
	/* try the provided URL first */
	request->url = URL;

	/* First request, open mode, only HEAD */
	if (request->open)
		request->op = "HEAD";
	else
		request->op = op;

	request->us = us;
	request->purl = purl;
	request->flags = flags;
	request->e = HTTP_PROTOCOL_ERROR;
	request->need_auth = 0;
	request->size = -1;

	if (request->direct && request->purl) {
		fetchFreeURL(request->purl);
		request->purl = NULL;
	}

	/* if the A flag is set, we only get one try */
	n = request->noredirect ? 1 : MAX_REDIRECT;
	i = 0;

	do {
		status = _http_single_request(request);
		if (status < 0)
			goto ouch;
		else if (status > 0)
			break;

		if (request->url != URL)
			fetchFreeURL(request->url);
		request->url = request->new;
		request->new = NULL;
	} while (++i < n);

	/* we failed, or ran out of retries */
	if (request->conn == NULL) {
		_http_seterr(request->e);
		goto ouch;
	}

	if (request->open)
		request->op = op;

	RMDBGLOG((HTTPDEBUG, "offset %lld, length %lld,"
		  " size %lld, clength %lld\n",
		  (RMint64)request->offset, (RMint64)request->length,
		  (RMint64)request->size, (RMint64)request->clength));

	/* check for inconsistencies, computed length and clength should be the same if set */
	if (request->clength != -1 && request->length != -1 && request->clength != request->length) {
		_http_seterr(HTTP_PROTOCOL_ERROR);
		goto ouch;
	}

	/* Content-Length: not provided, use computed length if we had a Range: */
	if (request->clength == -1 && request->length != -1)
		request->clength = request->length;

	/* We allow request->length != request->size, length is length of this
	 * request, whereas size is the total length of the target */
	/* if (request->length != -1 && request->size != -1 && request->length != request->size) {
		_http_seterr(HTTP_PROTOCOL_ERROR);
		goto ouch;
	}*/

	/* For a closed connection, use the content length as the size */
	if (request->size == -1 && !request->open)
		request->size = request->clength;

	/* fill in stats */
	if (request->us) {
		request->us->size = request->size;
		request->us->atime = request->us->mtime = request->mtime;
	}

	/* too far? */
	if (URL->offset > 0 && request->offset > URL->offset) {
		_http_seterr(HTTP_PROTOCOL_ERROR);
		goto ouch;
	}

	/* report back real offset and size */
	URL->offset = request->offset;
	URL->length = request->clength;

	/* wrap it up in a HTTPFile*/
	if ((f = _http_funopen(request)) == NULL) {
		_fetch_syserr();
		goto ouch;
	}

	/*if (request->url != URL)
		fetchFreeURL(request->url);
	if (request->purl)
		fetchFreeURL(request->purl);*/

	if (HTTP_ERROR(request->conn->err)) {
		fetchClose(f);
		f = NULL;
	}

	return (f);

ouch:
	if (request->cache)
		destroy_cache(request->cache);
	if (request->url)
		fetchFreeURL(request->url);
	if (request->purl)
		fetchFreeURL(request->purl);
	if (request->conn != NULL)
		_fetch_close(request->conn);
	if (request->host != NULL)
		RMFree(request->host);
	if (request->custom_header != NULL)
		RMFree(request->custom_header);
	RMFree(request);
	return (NULL);
}


/*****************************************************************************
 * Entry points
 */
/**
 * Set custom hooks. The custom hooks will be used if the
 * RM_HTTP_CUSTOM_HOOKS flag is set during the fetchOpen.
 *
 * @param cookie - cookie used by the hooks.
 * @param hooks - set of custom callbacks that use the cookie.
 * @return void
 */
void fetchSetCustomHooks(void *cookie, HttpHookOps *hooks)
{
	fetchCustomCookie = cookie;
	fetchCustomHooks = hooks;
}

/**
 * Set custom headers. The custom headers will be used if the
 * RM_HTTP_CUSTOM_HEADER flag is set during the fetchOpen.
 *
 * @param header - custom header, should use "\r\n" as end of line, NULL
 *                 terminated. The final "\r\n" will be added upon sending over
 *                 the wire.  This string will be duplicated by each fetchOpen,
 *                 so it should ne be freed before the fetchOpen.
 * @return void
 */
void fetchSetCustomHeader(RMascii *header)
{
	fetchCustomHeader = header;
}


/*
 * Retrieve and stat a file by HTTP
 */
HTTPFile *
fetchXGetHTTP(struct url *URL, struct url_stat *us, RMHTTPFlags flags)
{
	return (_http_request(URL, "GET", us, _http_get_proxy(flags), flags));
}

/*
 * Retrieve a file by HTTP
 */
HTTPFile *
fetchGetHTTP(struct url *URL, RMHTTPFlags flags)
{
	return (fetchXGetHTTP(URL, NULL, flags));
}

HTTPFile *
fetchOpen(const RMascii *URL, RMHTTPFlags flags)
{
	struct url *url;

	RMDBGLOG((ENABLE,"Open %s\n", URL));
	url = fetchParseURL(URL);
	if (url == NULL)
		return(NULL);
	
	return(fetchXGetHTTP(url, NULL, flags));
}

/*
 * Get an HTTP document's metadata
 */
RMint32
fetchStatHTTP(struct url *URL, struct url_stat *us, RMHTTPFlags flags)
{
	HTTPFile *f;

	f = _http_request(URL, "HEAD", us, _http_get_proxy(flags), flags);
	if (f == NULL)
		return (-1);
	fetchClose(f);
	return (0);
}

