/*
 * jdapistd.c
 *
 * Copyright (C) 1994-1996, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains application interface code for the decompression half
 * of the JPEG library.  These are the "standard" API routines that are
 * used in the normal full-decompression case.  They are not used by a
 * transcoding-only application.  Note that if an application links in
 * jpeg_start_decompress, it will end up linking in the entire decompressor.
 * We thus must separate this file from jdapimin.c to avoid linking the
 * whole decompression library into a transcoder.
 */
#ifdef __CC_ARM
#include <alloca.h>        //for alloca(...)
#else
#include <malloc.h>         //for alloca(...)
#endif

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"


/* Forward declarations */
LOCAL(boolean) output_pass_setup JPP((j_decompress_ptr cinfo));

//#define FRAME_WIDTH (720) //define somewhere in .h file !!!!!!!!!!!
//#define FRAME_HEIGHT (480)//define somewhere in .h file !!!!!!!!!!!

#define Y_PLANE 0
#define U_PLANE 1
#define V_PLANE 2

/*
 * Decompression initialization.
 * jpeg_read_header must be completed before calling this.
 *
 * If a multipass operating mode was selected, this will do all but the
 * last pass, and thus may take a great deal of time.
 *
 * Returns FALSE if suspended.  The return value need be inspected only if
 * a suspending data source is used.
 */

GLOBAL(boolean)
jpeg_start_decompress (j_decompress_ptr cinfo)
{
  if (cinfo->global_state == DSTATE_READY) {
    /* First call: initialize master control, select active modules */
    jinit_master_decompress(cinfo);
    if (cinfo->buffered_image) {
      /* No more work here; expecting jpeg_start_output next */
      cinfo->global_state = DSTATE_BUFIMAGE;
      return TRUE;
    }
    cinfo->global_state = DSTATE_PRELOAD;
  }
  if (cinfo->global_state == DSTATE_PRELOAD) {
    /* If file has multiple scans, absorb them all into the coef buffer */
    if (cinfo->inputctl->has_multiple_scans) {
#ifdef D_MULTISCAN_FILES_SUPPORTED
      for (;;) {
	int retcode;
	/* Call progress monitor hook if present */
	if (cinfo->progress != NULL)
	  (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
	/* Absorb some more input */
	retcode = (*cinfo->inputctl->consume_input) (cinfo);
	if (retcode == JPEG_SUSPENDED)
	  return FALSE;
	if (retcode == JPEG_REACHED_EOI)
	  break;
	/* Advance progress counter if appropriate */
	if (cinfo->progress != NULL &&
	    (retcode == JPEG_ROW_COMPLETED || retcode == JPEG_REACHED_SOS)) {
	  if (++cinfo->progress->pass_counter >= cinfo->progress->pass_limit) {
	    /* jdmaster underestimated number of scans; ratchet up one scan */
	    cinfo->progress->pass_limit += (long) cinfo->total_iMCU_rows;
	  }
	}
      }
#else
      ERREXIT(cinfo, JERR_NOT_COMPILED);
#endif /* D_MULTISCAN_FILES_SUPPORTED */
    }
    cinfo->output_scan_number = cinfo->input_scan_number;
  } else if (cinfo->global_state != DSTATE_PRESCAN)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  /* Perform any dummy output passes, and set up for the final pass */
  return output_pass_setup(cinfo);
}


/*
 * Set up for an output pass, and perform any dummy pass(es) needed.
 * Common subroutine for jpeg_start_decompress and jpeg_start_output.
 * Entry: global_state = DSTATE_PRESCAN only if previously suspended.
 * Exit: If done, returns TRUE and sets global_state for proper output mode.
 *       If suspended, returns FALSE and sets global_state = DSTATE_PRESCAN.
 */

LOCAL(boolean)
output_pass_setup (j_decompress_ptr cinfo)
{
  if (cinfo->global_state != DSTATE_PRESCAN) {
    /* First call: do pass setup */
    (*cinfo->master->prepare_for_output_pass) (cinfo);
    cinfo->output_scanline = 0;
    cinfo->global_state = DSTATE_PRESCAN;
  }
  /* Loop over any required dummy passes */
  while (cinfo->master->is_dummy_pass) {
#ifdef QUANT_2PASS_SUPPORTED
    /* Crank through the dummy pass */
    while (cinfo->output_scanline < cinfo->output_height) {
      JDIMENSION last_scanline;
      /* Call progress monitor hook if present */
      if (cinfo->progress != NULL) {
	cinfo->progress->pass_counter = (long) cinfo->output_scanline;
	cinfo->progress->pass_limit = (long) cinfo->output_height;
	(*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
      }
      /* Process some data */
      last_scanline = cinfo->output_scanline;
      (*cinfo->main->process_data) (cinfo, (JSAMPARRAY) NULL,
				    &cinfo->output_scanline, (JDIMENSION) 0);
      if (cinfo->output_scanline == last_scanline)
	return FALSE;		/* No progress made, must suspend */
    }
    /* Finish up dummy pass, and set up for another one */
    (*cinfo->master->finish_output_pass) (cinfo);
    (*cinfo->master->prepare_for_output_pass) (cinfo);
    cinfo->output_scanline = 0;
#else
    ERREXIT(cinfo, JERR_NOT_COMPILED);
#endif /* QUANT_2PASS_SUPPORTED */
  }
  /* Ready for application to drive output pass through
   * jpeg_read_scanlines or jpeg_read_raw_data.
   */
  cinfo->global_state = cinfo->raw_data_out ? DSTATE_RAW_OK : DSTATE_SCANNING;
  return TRUE;
}


/*
 * Read some scanlines of data from the JPEG decompressor.
 *
 * The return value will be the number of lines actually read.
 * This may be less than the number requested in several cases,
 * including bottom of image, data source suspension, and operating
 * modes that emit multiple scanlines at a time.
 *
 * Note: we warn about excess calls to jpeg_read_scanlines() since
 * this likely signals an application programmer error.  However,
 * an oversize buffer (max_lines > scanlines remaining) is not an error.
 */

GLOBAL(JDIMENSION)
jpeg_read_scanlines (j_decompress_ptr cinfo, JSAMPARRAY scanlines,
		     JDIMENSION max_lines)
{
  JDIMENSION row_ctr;

  if (cinfo->global_state != DSTATE_SCANNING)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  if (cinfo->output_scanline >= cinfo->output_height) {
    WARNMS(cinfo, JWRN_TOO_MUCH_DATA);
    return 0;
  }

  /* Call progress monitor hook if present */
  if (cinfo->progress != NULL) {
    cinfo->progress->pass_counter = (long) cinfo->output_scanline;
    cinfo->progress->pass_limit = (long) cinfo->output_height;
    (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
  }

  /* Process some data */
  row_ctr = 0;
  (*cinfo->main->process_data) (cinfo, scanlines, &row_ctr, max_lines);
  cinfo->output_scanline += row_ctr;
  return row_ctr;
}


/*
 * Alternate entry point to read raw data.
 * Processes exactly one iMCU row per call, unless suspended.
 */

GLOBAL(JDIMENSION)
jpeg_read_raw_data (j_decompress_ptr cinfo, JSAMPIMAGE data,
		    JDIMENSION max_lines)
{
  JDIMENSION lines_per_iMCU_row;

  if (cinfo->global_state != DSTATE_RAW_OK)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  if (cinfo->output_scanline >= cinfo->output_height) {
    WARNMS(cinfo, JWRN_TOO_MUCH_DATA);
    return 0;
  }

  /* Call progress monitor hook if present */
  if (cinfo->progress != NULL) {
    cinfo->progress->pass_counter = (long) cinfo->output_scanline;
    cinfo->progress->pass_limit = (long) cinfo->output_height;
    (*cinfo->progress->progress_monitor) ((j_common_ptr) cinfo);
  }

  /* Verify that at least one iMCU row can be returned. */
  lines_per_iMCU_row = cinfo->max_v_samp_factor * cinfo->min_DCT_scaled_size;
  if (max_lines < lines_per_iMCU_row)
    ERREXIT(cinfo, JERR_BUFFER_SIZE);

  /* Decompress directly into user's buffer. */
  if (! (*cinfo->coef->decompress_data) (cinfo, data))
    return 0;			/* suspension forced, can do nothing more */

  /* OK, we processed one iMCU row. */
  cinfo->output_scanline += lines_per_iMCU_row;
  return lines_per_iMCU_row;
}

#ifdef  YYYY_UV
GLOBAL(JDIMENSION) jpeg_read_image_420 (j_decompress_ptr cinfo, JSAMPLE *Y_plane, JSAMPLE *UV_plane)
{
#define COLOR_COMPONENTS 3
  int                 i;
  JDIMENSION          num_scanlines;
  JDIMENSION          plane_height[COLOR_COMPONENTS];
  JDIMENSION          plane_width[COLOR_COMPONENTS];
  int                 H_samp[COLOR_COMPONENTS];
  int                 V_samp[COLOR_COMPONENTS];
  
  JSAMPLE             *UV_ptr = NULL;
  JSAMPARRAY          buff[COLOR_COMPONENTS];
  JSAMPROW            plane_buff[COLOR_COMPONENTS];
  
  JDIMENSION          lines_per_iMCU_row = cinfo->max_v_samp_factor * cinfo->min_DCT_scaled_size;
  JDIMENSION          center_ajust_h = 0; // (((FRAME_WIDTH  - cinfo->cur_comp_info[Y_PLANE]->downsampled_width)>>1)&~0x7);
  JDIMENSION          center_ajust_v = 0; // FRAME_WIDTH*(((FRAME_HEIGHT - cinfo->cur_comp_info[Y_PLANE]->downsampled_height)>>1)&~0x7);
  

/*  JDIMENSION    yrow_width  = plane_width[Y_PLANE];
  JDIMENSION    yrow_height = plane_height[Y_PLANE];
  JDIMENSION    urow_width  = plane_width[U_PLANE];
  JDIMENSION    urow_height = plane_height[U_PLANE];
  JDIMENSION    vrow_width  = plane_width[V_PLANE];
  JDIMENSION    vrow_height = plane_height[V_PLANE];
*/
  
  for(i=0; i<cinfo->comps_in_scan; i++)
  {
    buff[i]       = NULL;
    plane_buff[i] = NULL;//keep pointers for free(...) 
    plane_height[i]  = cinfo->cur_comp_info[i]->downsampled_height;
    plane_width[i]   = cinfo->cur_comp_info[i]->downsampled_width;
    
    H_samp[i]= cinfo->cur_comp_info[i]->DCT_scaled_size * cinfo->cur_comp_info[i]->h_samp_factor;
    V_samp[i]= cinfo->cur_comp_info[i]->DCT_scaled_size * cinfo->cur_comp_info[i]->v_samp_factor;

    //XXX

  }

    // padd size to next DCT block
//  plane_width[i] = ((plane_width[i] - 1)/cinfo->cur_comp_info[i]->DCT_scaled_size + 1) * cinfo->cur_comp_info[i]->DCT_scaled_size;
//  plane_height[i] = ((plane_height[i] - 1)/cinfo->cur_comp_info[i]->DCT_scaled_size + 1) * cinfo->cur_comp_info[i]->DCT_scaled_size;
    // padd size to next  16x16 block
  plane_width[Y_PLANE] = ((plane_width[Y_PLANE] - 1)/16 + 1) * 16;
  plane_height[Y_PLANE] = ((plane_height[Y_PLANE] - 1)/16 + 1) * 16;


  for(i=1; i<3; i++) {
    if (i < cinfo->num_components) {
       plane_width[i] = plane_width[Y_PLANE]*H_samp[i]/H_samp[Y_PLANE];
	   plane_height[i] = plane_height[Y_PLANE]*V_samp[i]/V_samp[Y_PLANE];
	} else {
       plane_width[i] = plane_width[Y_PLANE];
	   plane_height[i] = plane_height[Y_PLANE];
	}
  }

#if 0  
  for(i=0; i<cinfo->num_components; i++) {
	printf("XXX Plane %i : dssz=%dx%d, dctsz=%d, plane=%dx%d, sf=%dx%x, adjsf=%dx%d\n",i, 
		cinfo->cur_comp_info[i]->downsampled_width,
	    cinfo->cur_comp_info[i]->downsampled_height,
	    cinfo->cur_comp_info[i]->DCT_scaled_size,
	    plane_width[i], plane_height[i],
	    cinfo->cur_comp_info[i]->h_samp_factor,
	    cinfo->cur_comp_info[i]->v_samp_factor,
	    H_samp[i], V_samp[i]);

  }

#endif
  
  ///////
  plane_buff[Y_PLANE]  = Y_plane;
  for(i=1; i<cinfo->num_components;i++) {
    plane_buff[i] = (JSAMPROW)
      (*cinfo->mem->alloc_large)((j_common_ptr) cinfo, JPOOL_IMAGE,
				 plane_width[i]*plane_height[i]* 
				 sizeof(JSAMPLE));
  }
  {
    JDIMENSION      j;
    JSAMPROW        rowptr        = plane_buff[Y_PLANE] + center_ajust_h + center_ajust_v;
    JDIMENSION      lines_in_row  = cinfo->cur_comp_info[Y_PLANE]->v_samp_factor * cinfo->cur_comp_info[Y_PLANE]->DCT_scaled_size;
    
    buff[Y_PLANE] = alloca(lines_in_row*sizeof(JSAMPARRAY));
    for(j=0; j<lines_in_row; j++) {
      buff[Y_PLANE][j]  = rowptr;
      rowptr     += plane_width[Y_PLANE];
    }
  }

  for(i=1; i<cinfo->num_components;i++) {
    JDIMENSION      j;
    JSAMPROW        rowptr        = plane_buff[i];
    JDIMENSION      lines_in_row  = cinfo->cur_comp_info[i]->v_samp_factor * cinfo->cur_comp_info[i]->DCT_scaled_size;
    
    buff[i] = alloca(lines_in_row*sizeof(JSAMPARRAY));
    for(j=0; j<lines_in_row; j++) {
      buff[i][j]  = rowptr;
      rowptr     += plane_width[i];
    }
  } 
  
    /* Process data */
  while (cinfo->output_scanline < cinfo->output_height) {
    JDIMENSION      j;
    JDIMENSION      lines_in_row;

    num_scanlines =  jpeg_read_raw_data(cinfo, buff, lines_per_iMCU_row);
    lines_in_row  = cinfo->cur_comp_info[Y_PLANE]->v_samp_factor * cinfo->cur_comp_info[Y_PLANE]->DCT_scaled_size;
    for(j=0; j<lines_in_row; j++) {
      buff[Y_PLANE][j]  += lines_in_row*plane_width[Y_PLANE];
    }

    for(i=1; i<cinfo->num_components;i++) {
      lines_in_row  = cinfo->cur_comp_info[i]->v_samp_factor * cinfo->cur_comp_info[i]->DCT_scaled_size;
      for(j=0; j<lines_in_row; j++) {
       buff[i][j]  += lines_in_row*plane_width[i];
      }
    }
  }
  
  { //convert U & V planes to UVUV plane 4:2:0
    JDIMENSION    i;

    JSAMPROW      urow0 = plane_buff[U_PLANE];
    JSAMPROW      urow1 = urow0 + plane_width[U_PLANE];
    JSAMPROW      vrow0 = plane_buff[V_PLANE];
    JSAMPROW      vrow1 = vrow0 + plane_width[V_PLANE];
     
    UV_ptr = UV_plane + center_ajust_h + (center_ajust_v>>1);
    
    if (cinfo->num_components < 3) {  // case Y1 (4:0:0) black&white
      for(i=0; i<plane_height[U_PLANE];i+=2) {
        JDIMENSION j;
        for(j=0; j<plane_width[U_PLANE]; j+=2) {
          *UV_ptr++ = 0x80;
          *UV_ptr++ = 0x80;
        }
      }
    } else 
    if(H_samp[Y_PLANE]==H_samp[U_PLANE] &&
       H_samp[Y_PLANE]==H_samp[V_PLANE] &&
       H_samp[Y_PLANE]==V_samp[U_PLANE] &&
       V_samp[Y_PLANE]==V_samp[V_PLANE]
       )
    { // case Y1:U1:V1 (4:4:4)
//      cinfo->output_width&=~1;//deleteme 
//      urow_width = (urow_width + urow_width - cinfo->output_width)*sizeof(JSAMPLE);
//      vrow_width = (vrow_width + vrow_width - cinfo->output_width)*sizeof(JSAMPLE);
      for(i=0; i<plane_height[U_PLANE];i+=2)
      {
        JDIMENSION j;
        for(j=0; j<plane_width[U_PLANE]; j+=2)
        { //calc average U & V from 4 to 1
         unsigned char a, b, c, d;
         a = *urow0; 
         urow0++;
         b = *urow0;
         urow0++;
         c = *urow1; 
         urow1++;
         d = *urow1;
         urow1++;
         *UV_ptr++ = (a + b + c + d)>>2;
         a = *vrow0; 
         vrow0++;
         b = *vrow0;
         vrow0++;
         c = *vrow1; 
         vrow1++;
         d = *vrow1;
         vrow1++;
         *UV_ptr++ = (a + b + c + d)>>2;
//         *UV_ptr++ = (*urow0++ + *urow0++ + *urow1++ + *urow1++)>>2;
//         *UV_ptr++ = (*vrow0++ + *vrow0++ + *vrow1++ + *vrow1++)>>2;
        }
        urow0+= plane_width[U_PLANE];
        urow1+= plane_width[U_PLANE];
        vrow0+= plane_width[U_PLANE];
        vrow1+= plane_width[U_PLANE];
      }
    } else if( H_samp[Y_PLANE]==2*H_samp[U_PLANE] &&
	       H_samp[Y_PLANE]==2*H_samp[V_PLANE] &&
	       H_samp[Y_PLANE]==2*V_samp[U_PLANE] &&
	       V_samp[Y_PLANE]==2*V_samp[V_PLANE])
    {// case Y4:U1:V1 (4:2:0)
      for(i=0; i<plane_height[U_PLANE];i++)
      {
        JDIMENSION j;
        for(j=0; j<plane_width[U_PLANE]; j++)
        { // jast mix U & V 
          *UV_ptr++ = *urow0++;
          *UV_ptr++ = *vrow0++;  
        }
      }
    }else if(H_samp[Y_PLANE]==2*H_samp[U_PLANE] &&
	     H_samp[Y_PLANE]==2*H_samp[V_PLANE] &&
	     V_samp[Y_PLANE]==V_samp[U_PLANE] &&
	     V_samp[Y_PLANE]==V_samp[V_PLANE])
    {// case YY:U:V (4:2:2)
      for(i=0; i<plane_height[U_PLANE];i+=2)
	{
        JDIMENSION j;
        for(j=0; j<plane_width[U_PLANE]; j++)
        { //calc average U & V from 4 to 1
          *UV_ptr++ = (*urow0++ + *urow1++)>>1;
          *UV_ptr++ = (*vrow0++ + *vrow1++)>>1;  
        }
        urow0+= plane_width[U_PLANE];
        urow1+= plane_width[U_PLANE];
        vrow0+= plane_width[V_PLANE]; 
        vrow1+= plane_width[V_PLANE];
      }
    } else if( H_samp[Y_PLANE]==H_samp[U_PLANE] &&
	       H_samp[Y_PLANE]==H_samp[V_PLANE] &&
	       V_samp[Y_PLANE]==2*V_samp[U_PLANE] &&
	       V_samp[Y_PLANE]==2*V_samp[V_PLANE])
     {// case Y:U:V  (4:2:2 rotated)
      //      Y: 
      for(i=0; i<plane_height[U_PLANE];i++)
      {
        JDIMENSION j;
        for(j=0; j<plane_width[U_PLANE]; j+=2)
        { //calc average U & V from 2 to 1
         unsigned char a, b;
         a = *urow0; 
         urow0++;
         b = *urow0;
         urow0++;
         *UV_ptr++ = (a + b)>>1;
         a = *vrow0; 
         vrow0++;
         b = *vrow0;
         vrow0++;
         *UV_ptr++ = (a + b)>>1;
//          *UV_ptr++ = (*urow0++ + *urow0++)>>1;
//          *UV_ptr++ = (*vrow0++ + *vrow0++)>>1;  
        }
      }
    } else {
      printf("\n**** Usupported format****\n");
      printf("H_samp[Y]  == %d H_samp[U]  == %d H_samp[V]  == %d\n", H_samp[Y_PLANE], H_samp[U_PLANE], H_samp[V_PLANE]);
      printf("V_samp[Y]  == %d V_samp[U]  == %d V_samp[V]  == %d\n", V_samp[Y_PLANE], V_samp[U_PLANE], V_samp[V_PLANE]);

    }
  }
  
  return cinfo->output_scanline;
}
#endif //YYYY_UV

/* Additional entry points for buffered-image mode. */

#ifdef D_MULTISCAN_FILES_SUPPORTED

/*
 * Initialize for an output pass in buffered-image mode.
 */

GLOBAL(boolean)
jpeg_start_output (j_decompress_ptr cinfo, int scan_number)
{
  if (cinfo->global_state != DSTATE_BUFIMAGE &&
      cinfo->global_state != DSTATE_PRESCAN)
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  /* Limit scan number to valid range */
  if (scan_number <= 0)
    scan_number = 1;
  if (cinfo->inputctl->eoi_reached &&
      scan_number > cinfo->input_scan_number)
    scan_number = cinfo->input_scan_number;
  cinfo->output_scan_number = scan_number;
  /* Perform any dummy output passes, and set up for the real pass */
  return output_pass_setup(cinfo);
}


/*
 * Finish up after an output pass in buffered-image mode.
 *
 * Returns FALSE if suspended.  The return value need be inspected only if
 * a suspending data source is used.
 */

GLOBAL(boolean)
jpeg_finish_output (j_decompress_ptr cinfo)
{
  if ((cinfo->global_state == DSTATE_SCANNING ||
       cinfo->global_state == DSTATE_RAW_OK) && cinfo->buffered_image) {
    /* Terminate this pass. */
    /* We do not require the whole pass to have been completed. */
    (*cinfo->master->finish_output_pass) (cinfo);
    cinfo->global_state = DSTATE_BUFPOST;
  } else if (cinfo->global_state != DSTATE_BUFPOST) {
    /* BUFPOST = repeat call after a suspension, anything else is error */
    ERREXIT1(cinfo, JERR_BAD_STATE, cinfo->global_state);
  }
  /* Read markers looking for SOS or EOI */
  while (cinfo->input_scan_number <= cinfo->output_scan_number &&
	 ! cinfo->inputctl->eoi_reached) {
    if ((*cinfo->inputctl->consume_input) (cinfo) == JPEG_SUSPENDED)
      return FALSE;		/* Suspend, come back later */
  }
  cinfo->global_state = DSTATE_BUFIMAGE;
  return TRUE;
}

#endif /* D_MULTISCAN_FILES_SUPPORTED */
