/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#define ALLOW_OS_CODE 1
#include "rmdef/rmdef.h"
#include "rmcore/include/rmcore.h"
#include "../include/ms_cardea_types.h"
#include "wmdrmndcoreapi.h"
#include "../include/ms_cardea.h"

#if 0
#define CARDEADBG ENABLE
#else
#define CARDEADBG DISABLE
#endif

/* For testing create a fake device for any url in init_carde */
/* #define TEST_FAKE_DEVICE */

#ifdef TEST_FAKE_DEVICE
extern wmdrmnet_context * init_wmdrmnet_context_fake(RMuint8 *content_key);
/* Harcoded key for test purpose*/
static RMuint8 content_key[] = { /* 16 bytes */
	0x58, 0xfa, 0x13, 0x0e, 0xef, 0x29, 0x19, 0xfe,
	0x26, 0x7f, 0xb6, 0xe1, 0xc8, 0x9d, 0x21, 0x66 
};
#endif

/** All the knwon ms_device, keep track of them, also used by ms_cardea.h, so not static */
struct ms_upnp_extension_s *ms_device[MAX_MS_DEVICE]; 

/**
 * Initialize CARDEA for a given URL
 *
 * This function tries to get the CARDEA session associated with the URL. If no
 * session is opened for this URL, returns NULL.
 *
 * @param url - url for which a session header is required
 * @return NULL if no session header is available
 */
void * init_cardea(const RMascii *url)
{
	RMuint32 i;
	RMstatus status;

#ifdef TEST_FAKE_DEVICE
	for (i=0;i<MAX_MS_DEVICE;i++){
		if (ms_device[i] == NULL){
			RMDBGLOG((ENABLE,"Create a fake device for %s : index = %lu\n", url, i));
			ms_device[i] = RMMalloc(sizeof(struct ms_upnp_extension_s));
			ms_device[i]->flags = 0;
			ms_device[i]->drm_context = init_wmdrmnet_context_fake(content_key);
			if (ms_device[i]->drm_context == NULL)
				RMDBGLOG((ENABLE,"Error creating DRM context for fake device %s\n", url));
			ms_device[i]->session_header = RMMallocAndDuplicateAscii("User-Agent: Fake device");
			ms_device[i]->url = RMMallocAndDuplicateAscii(url);
			break;
		}
	}
#endif

	
	for (i=0;i<MAX_MS_DEVICE;i++){
		if (ms_device[i] != NULL &&
		    ms_device[i]->url != NULL &&
		    RMCompareAscii(url, ms_device[i]->url)){
			/* Clear MS stream context */
			RMMemset(ms_device[i]->ms_stream_decryption_context, 0, sizeof(ms_device[i]->ms_stream_decryption_context));
			/* Initialize CARDEA core */
			status = init_cardea_core(ms_device[i]->drm_context);
			if (RMFAILED(status)){
				RMDBGLOG((ENABLE,"Cannot initialize CARDEA core\n"));
				return NULL;
			}
			return (void *)ms_device[i];
		}
	}

	return NULL;
}


/**
 * Terminate CARDEA
 *
 * This function handles the termination of the CARDEA session, if any.
 * 
 * @return RM_OK on success
 */
RMstatus term_cardea()
{
	/* Terminate CARDEA core */
	return term_cardea_core();
}


/**
 * Update the sample ID for a particuler stream and media object
 *
 * @param cardea_context - the context to update
 * @param stream_index - current stream index
 * @param media_object_number - current media object number
 * @param sample_id - pointer to the sample ID data
 * @param sample_id_size - size of the sample id (should be 8)
 */
void update_cardea_sample_id(
		void * cardea_context, 
		RMuint16 stream_index, 
		RMuint32 media_object_number, 
		RMuint8 *sample_id, 
		RMuint32 sample_id_size)
{
	struct ms_upnp_extension_s *ms_context = (struct ms_upnp_extension_s *) cardea_context;

	if (ms_context == NULL){
		RMDBGLOG((ENABLE,"** Error ** NULL context\n"));
		return;
	}

	if (stream_index == 0 || stream_index > 127){
		RMDBGLOG((ENABLE,"** Error ** stream index value is invalid : %ld\n", stream_index));
		return;
	}

	if (sample_id_size != 8)
		RMDBGLOG((ENABLE,"** Warning ** sample Id size is not 8 (%ld)\n", sample_id_size));

	if (ms_context->ms_stream_decryption_context[stream_index].media_object_number != media_object_number){
		ms_context->ms_stream_decryption_context[stream_index].media_object_number = media_object_number;
		RMMemset(ms_context->ms_stream_decryption_context[stream_index].sample_id, 0, 8);
		RMMemcpy(ms_context->ms_stream_decryption_context[stream_index].sample_id, sample_id, sample_id_size);
		/* Reset the byte count since we change media object */
		ms_context->ms_stream_decryption_context[stream_index].byte_count = 0;	
		RMDBGLOG((CARDEADBG,"ST:%02d,MON:%04lu,SID:0x%08lx%08lx\n", stream_index, media_object_number,
					RMbeBufToUint32(ms_context->ms_stream_decryption_context[stream_index].sample_id),
					RMbeBufToUint32(ms_context->ms_stream_decryption_context[stream_index].sample_id+4)));
					
	}

	/* If we are here, it means the url associated with the CARDEA context
	 * has been or is currently being retrieved, so the license is not
	 * valid anymore, reset the URL and session header fields */
	if (ms_context->url != NULL){
		RMFree(ms_context->url);
		ms_context->url = NULL;
	}

	if (ms_context->session_header != NULL){
		RMFree(ms_context->session_header);
		ms_context->session_header = NULL;
	}
}


/**
 * Decrypt cardea encrypted data (in place)
 *
 * @param cardea_context - the current cardea context
 * @param stream_index - current stream index
 * @param media_object_number - current media object number
 * @param buf - pointer to the data to decrypt
 * @param size - size of the data to decrypt
 * @return RM_OK on success
 *
 */
RMstatus decrypt_cardea_sample(
		void * cardea_context,
		RMuint16 stream_index,
		RMuint32 media_object_number,
		RMuint8 *buf,
		RMuint32 size)
{
	struct ms_upnp_extension_s *ms_context = (struct ms_upnp_extension_s *) cardea_context;
	RMstatus status;
	
	if (ms_context == NULL){
		RMDBGLOG((ENABLE,"** Error ** NULL context\n"));
		return RM_ERROR;
	}

	if (stream_index == 0 || stream_index > 127){
		RMDBGLOG((ENABLE,"** Error ** stream index value is invalid : %ld\n", stream_index));
		return RM_ERROR;
	}

	/* Check the media object number is the same as the one registered with the sample ID */
	if (ms_context->ms_stream_decryption_context[stream_index].media_object_number != media_object_number){
		RMDBGLOG((ENABLE,"** Errror ** media object number mismatch : %lu != %lu\n",
					media_object_number,
					ms_context->ms_stream_decryption_context[stream_index].media_object_number));
		return RM_ERROR;
	}

	status = decrypt_cardea_sample_core(
			ms_context->drm_context,
			ms_context->ms_stream_decryption_context[stream_index].sample_id,
			buf,
			size,
			ms_context->ms_stream_decryption_context[stream_index].byte_count);
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"** Error ** CARDEA core decryption failed.\n"));
		return status;
	}

	/* Update the byte count */
	ms_context->ms_stream_decryption_context[stream_index].byte_count += size;

	return RM_OK;
}


/**
 * Get the output protection level for a given cardea context
 *
 * @param cardea_context - the current cardea context
 * @param opl - pointer to the structure that should be
 *    filled with the output protection level data
 * @return RM_OK on success
 */
RMstatus get_cardea_opl(
		void * cardea_context,
		struct cardea_opl_s *opl)
{
	struct ms_upnp_extension_s *ms_context = (struct ms_upnp_extension_s *) cardea_context;

	if (ms_context == NULL || ms_context->drm_context == NULL || opl == NULL)
		return RM_ERROR;

	/* Could be more than a simple cast, don't rely on it */
	return get_cardea_opl_core(ms_context->drm_context, (struct cardea_opl_core_s *) opl);
}


/**
 * Get the session ID header for a given URL
 *
 * This function tries to get the session header associated with a URL. If no
 * MS DRM session header is required, or if a license has not been obtained,
 * this function will return NUL.
 *
 * @param url - url for which a session header is required
 * @return NULL if no session header is available
 */
RMascii * get_ms_session_header(const RMascii *url)
{
	RMuint32 i;
	
	for (i=0;i<MAX_MS_DEVICE;i++){
		if (ms_device[i] != NULL &&
		    ms_device[i]->url != NULL &&
		    RMCompareAscii(url, ms_device[i]->url) &&
		    ms_device[i]->session_header)
			return ms_device[i]->session_header;
	}

	return NULL;
}




