/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#define ALLOW_OS_CODE 1
#include "rmdef/rmdef.h"
#include "../include/ms_upnp.h"
#include "ms_upnp_types.h"
#include "ms_upnp_event.h"
#include "wmdrmndcoreapi.h"
#include "ms_license.h"
#include "ms_tcp.h"
#include "rmcore/include/rmascii.h"
#include <stdio.h>
#ifdef CARDEA_TEST_TIMING
#include <sys/time.h>
#include <time.h>
#endif /* CARDEA_TEST_TIMING */

static RMascii request_template[] = 
	"POST %s HTTP/1.0\r\n"
	"Accept: */*\r\n"
	"Supported: com.microsoft.wmdrm-nd\r\n"
	"Content-Type: application/vnd.ms-wmdrm-license-request\r\n"
	"Content-Length: %ld\r\n"
	"\r\n";

/* Maximum number of digit required to represent an unsigned 32 bits */
#define MAX_U32_STR 10

/* Timeout in s waiting for HTTP answer */
#define LICENSE_TIMEOUT 50

/* String to identify the session header in the HTTP answer */
#define HEADER_SESSION	"WMDRM-ND: SessionId="

/* String to identify error status */
#define WMDRM_ND_STATUS	"WMDRM-ND-Status:"

#ifdef CARDEA_TEST_TIMING
static void timing_info(struct timeval *tv1, struct timeval *tv2, RMascii *message)
{
	RMint32 tv_sec, tv_usec;

	tv_sec = tv2->tv_sec - tv1->tv_sec;
	tv_usec = tv2->tv_usec - tv1->tv_usec;
	if (tv_usec < 0) {
		tv_sec--;
		tv_usec = 1000000 + tv_usec;
	}
	printf("%s : %lu s %lu us\n", message, tv_sec, tv_usec);
}
#endif /* CARDEA_TEST_TIMING */

/**
 * Get next http header
 *
 * @param http - pointer to the message to parse (will be modified after each call)
 * @param http_size - size of the message (will be modified also)
 * @return RMascii * - pointer to the next header, NULL terminated
 */
static RMascii * get_header(RMuint8 **http, RMuint32 *http_size)
{
	RMuint32 i=0;
	RMuint8 *header = *http;

	for (i=0; (i+1)<(*http_size); i++){
		if ( ((*http)[i] == '\r') && ((*http)[i+1] == '\n') ){
			(*http)[i] = 0;
			*http += (i+2);
			if (*http_size >= (i+2)){
				*http_size -= (i+2);
			} else {
				*http_size = 0;
			}
			return (RMascii *) header;
		}
	}

	return NULL;
}
	
	
/**
 * Parse a HTTP response, returns a pointer to the content
 *
 * @param device - UPnP server we are working for
 * @param http - http response to parse
 * @param http_size - size of the http response
 * @param ms_error - MS error code returned in case of failure (see ms_upnp_types.h)
 * @return RM_OK on success
 */
static RMstatus parse_http(struct UPnPDevice *device, RMuint8 **http, RMuint32 *http_size, RMuint32 *ms_error)
{
	RMascii *header;

	if (ms_error == NULL)
		return RM_ERROR;

	*ms_error = 0;

	while (*http_size > 0){
		header = get_header(http, http_size);
		if (header == NULL) 
			/* Error */
			return RM_ERROR;
		if (header[0] == 0)
			/* End of headers, data is next */
			break;
		RMDBGLOG((ENABLE,"Got header : %s\n", header));
		if (RMNCompareAscii(header, HEADER_SESSION, RMasciiLength(HEADER_SESSION))){
			/* DRM Session ID, keep it for later */
			MS_UPNP(device)->session_header = RMMallocAndDuplicateAscii(header);
			RMDBGLOG((ENABLE,"  Saved session ID\n"));
		}else if (RMNCompareAscii(header, WMDRM_ND_STATUS, RMasciiLength(WMDRM_ND_STATUS))){
			if (RMasciiLength(header) > (RMasciiLength(WMDRM_ND_STATUS)+4)){
				header[RMasciiLength(WMDRM_ND_STATUS)+4] = 0;
				RMasciiToUInt32(header+RMasciiLength(WMDRM_ND_STATUS)+1, ms_error);
			}
		}
	}

	RMDBGLOG((ENABLE,"HTTP content : %ld bytes\n", *http_size));

	if (*ms_error)
		return RM_ERROR;

	return RM_OK;
}


/*
 * Get a license for a given URL, on the given device
 * @param device
 * @param uri - URL to get
 * @param session_header - returns the session header for this URL, should not
 *                         be freed or modified.
 * @return 0 on success, < 0 on error
 */
RMint32 get_ms_license(struct UPnPDevice *device, RMascii *url, RMascii **session_header)
{
	ms_tcp_socket_t *ms_socket = NULL;
	RMstatus status;
	RMint32 rc = 0;
	RMuint8 *message;
	RMuint32 message_size;
	RMuint8 *answer;
	RMuint32 answer_size;
	RMuint8 *http_request = NULL;
	RMuint32 http_request_size;
	RMint32 written;
	RMuint32 ms_error = 0;
#ifdef CARDEA_TEST_TIMING
	struct timeval tv1, tv2;
#endif /* CARDEA_TEST_TIMING */
	
	if (device == NULL || 
	    MS_UPNP(device) == NULL ||
	    MS_DRM_CONTEXT(device) == NULL ||
	    url == NULL){
		rc = -1;
		goto error;
	}

#ifdef CARDEA_TEST_TIMING
	gettimeofday(&tv1, NULL);
#endif /* CARDEA_TEST_TIMING */
	/* Get license request message */
	status = get_license_request_message(
			MS_DRM_CONTEXT(device),
			&message,
			&message_size);
#ifdef CARDEA_TEST_TIMING
	gettimeofday(&tv2, NULL);
	timing_info(&tv1, &tv2, "Generate license request");
#endif /* CARDEA_TEST_TIMING */
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"Error getting license request message\n"));
		rc = -1;
		goto error;
	} else {
		RMDBGLOG((ENABLE,"Got %ld bytes license request message.\n", message_size));
	}

	/* Create HTTP request */
#ifdef CARDEA_TEST_TIMING
	gettimeofday(&tv1, NULL);
#endif /* CARDEA_TEST_TIMING */
	http_request_size = RMasciiLength(request_template)+RMasciiLength(url)+MAX_U32_STR+message_size;
	http_request = (RMuint8 *)RMMalloc(http_request_size);
	if (http_request == NULL){
		RMDBGLOG((ENABLE,"Error, cannot allocate http_request\n"));
		rc = -1;
		goto error;
	}

	written = snprintf((RMascii *)http_request, (size_t) http_request_size, request_template, url, message_size);
	if ((unsigned)written >= http_request_size){
		RMDBGLOG((ENABLE,"Error, HTTP request too big.\n"));
		rc = -1;
		goto error;
	}

	RMMemcpy(http_request+written,message,message_size);
#ifdef CARDEA_TEST_TIMING
	gettimeofday(&tv2, NULL);
	timing_info(&tv1, &tv2, "Create HTTP request");

	gettimeofday(&tv1, NULL);
#endif /* CARDEA_TEST_TIMING */
	/* Open the connection */
	ms_socket = open_ms_tcp_socket(url);
	if (ms_socket == NULL){
		RMDBGLOG((ENABLE,"Error opening MS socket\n"));
		rc = -1;
		goto error;
	} else {
		RMDBGLOG((ENABLE,"Opened MS socket\n"));
	}

	/* Send the request, receive the answer */
	status = ms_tcp_send_receive(ms_socket, http_request, written+message_size, &answer, &answer_size, LICENSE_TIMEOUT);
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"Error sending license request\n"));
		rc = -1;
		goto error;
	}
#ifdef CARDEA_TEST_TIMING
	gettimeofday(&tv2, NULL);
	timing_info(&tv1, &tv2, "HTTP request (conn+send+receive)");

	gettimeofday(&tv1, NULL);
#endif /* CARDEA_TEST_TIMING */
	status = parse_http(device,&answer, &answer_size, &ms_error);
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"Error parsing HTTP : %ld\n", ms_error));
		rc = -ms_error;
		goto error;
	}

	status = process_license(MS_DRM_CONTEXT(device), answer, answer_size);
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"Error processing license\n"));
		rc = -1;
		goto error;
	}
#ifdef CARDEA_TEST_TIMING
	gettimeofday(&tv2, NULL);
	timing_info(&tv1, &tv2, "Parse HTTP response and process license");
#endif /* CARDEA_TEST_TIMING */

	*session_header = MS_UPNP(device)->session_header;
	MS_UPNP(device)->url = RMMallocAndDuplicateAscii(url);
	if (MS_UPNP(device)->url == NULL){
		RMDBGLOG((ENABLE,"Error RMMallocAndDuplicateAscii\n"));
		rc = -1;
		goto error;
	}

	RMDBGLOG((ENABLE,"License OK\n"));
error:
	if (ms_socket)
		close_ms_tcp_socket(ms_socket);

	if (http_request)
		RMFree(http_request);

	return rc;
}

