/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#define ALLOW_OS_CODE 1
#include "rmdef/rmdef.h"
#include "ms_upnp_event.h"
#include "../include/ms_upnp.h"
#include "wmdrmndcoreapi.h"
#include "MSCP_ControlPoint.h"
#include "UPnPControlPointStructs.h"
#include "ms_upnp_types.h"
#include <stdio.h>
#include "ms_udp.h"


/* Maximum number of proximity detection attempt */
#define MAX_PROX_ATTEMPT 50

static RMstatus test_ms_proximity(struct UPnPDevice *device);

#ifdef WITH_CARDEA_DIRECT_API

void init_ms_client(void);

/* Include should define serial_number, device_cert, and private_key */
#include "private_settings.inc"

/* 
 * Initialize client internals : load its key and certificate for example
 * XXX This is done by the deamon, if enabled.
 */
void init_ms_client(void)
{
	RMstatus status;

	status = init_receiver(private_key, sizeof(private_key), device_cert, sizeof(device_cert), serial_number);
	if (RMSUCCEEDED(status)){
		RMDBGLOG((ENABLE,"Device manufacture ok\n"));
	}else{
		RMDBGLOG((ENABLE,"Device manufacture failed\n"));
	}
}
#endif /* WITH_CARDEA_DIRECT_API */

/*
 * Callback for registration message response, called when the device receives the response from the MS server
 *
 * @param service - the service called
 * @param error - error code, != 0 on error
 * @param user - user data
 * @param message - received message
 * @param message_size
 */
static void callback_registration_message(struct UPnPService* service, int error, void *user, unsigned char *message, int message_size)
{
	struct UPnPDevice *device;
	RMstatus status;

	RMDBGLOG((ENABLE,"Received registration answer\n"));
	if (service != NULL){
		device = service->Parent;
	} else {
		RMDBGLOG((ENABLE,"Error, NULL service\n"));
		return;
	}
	
	if (error != 0) {
		fprintf(stderr,"Error in registration : %d on %s\n", error, device->FriendlyName);
	} else {
		RMDBGLOG((ENABLE,"Registration successful : %d, process the response ...\n", error));
		status = process_registration_response(MS_DRM_CONTEXT(device), message, message_size);
		if (RMSUCCEEDED(status)){
			RMDBGLOG((ENABLE, "Registration successful on %s, testing proximity ... \n", device->FriendlyName));
			MS_UPNP(device)->flags |= MS_ISVALIDATED;
			status = test_ms_proximity(device);
			if (RMFAILED(status)){
				fprintf(stderr,"Proximity detection failed on %s\n", device->FriendlyName);
			} else {
				fprintf(stderr,"Promixity detection ok on %s\n", device->FriendlyName);
			}
		} else {
			fprintf(stderr, "Registration failed on on %s\n", device->FriendlyName);
		}
	}
	/* Not waiting registration anymore */
	MS_UPNP(device)->flags &= ~MS_WAITREGISTRATION;
	/* If regisatretion was forced, don't do it again */
	MS_UPNP(device)->flags &= ~MS_FORCEREGISTER;
}

/*
 * Try to register the client to the given device
 * @param device - device to register to
 * @return < 0 on error
 */
RMint32 register_ms_client(struct UPnPDevice *device)
{
	struct UPnPService *service;
	RMuint8 *message;
	RMuint32 message_size;
	RMstatus status;

	service = MSCP_GetService_Registrar(device);
	if (service == NULL){
		RMDBGLOG((ENABLE,"Error : cannot register to %s, not WMC.\n",device->FriendlyName));
		return -1;
	}

	status = get_registration_message(MS_DRM_CONTEXT(device), &message, &message_size);
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"Error getting registration message fo %s\n", device->FriendlyName));
		return -1;
	}
	/* Set the new flags for this server : waiting validation */
	MS_UPNP(device)->flags |= MS_WAITREGISTRATION; 
	MSCP_Invoke_Registrar_RegisterDevice(service, callback_registration_message, (void *)NULL, message, message_size);
		
	/* Already registered, or registration sent, waiting for response */
	return 0;
}


/*
 * Try to perform proximity detection for the current client, with the given device as server
 * @param device - device to test proximity with
 * @return RM_OK on success, RM_ERROR else
 */
static RMstatus test_ms_proximity(struct UPnPDevice *device)
{
	RMuint8 *message;
	RMuint32 message_size;
	RMuint8 *answer;
	RMuint32 answer_size;
	RMstatus status;
	RMascii *address;
	ms_udp_socket_t *ms_socket;
	RMuint32 attempt = 0;
	
	if (device == NULL){
		RMDBGLOG((ENABLE,"Null device in test_ms_proximity\n"));
		return RM_ERROR;
	}

	if ((MS_UPNP(device)->flags & MS_ISVALIDATED) == 0){
		RMDBGLOG((ENABLE,"Device %s not registered, cannot process proximity detection.\n"));
		return RM_ERROR;
	}

	/* Get proximity address */
	status = get_proximity_address(MS_DRM_CONTEXT(device), &address);
	if (RMFAILED(status)){
		RMDBGLOG((ENABLE,"Error getting proximity address\n"));
		return RM_ERROR;
	}

	/* Open the socket */
	ms_socket = open_ms_udp_socket(address);
	if (ms_socket == NULL){
		RMDBGLOG((ENABLE,"Error opening MS socket\n"));
		goto error;
	} else {
		RMDBGLOG((ENABLE,"Opened MS socket\n"));
	}

	/* Update our state with this device */
	MS_UPNP(device)->flags |= MS_WAITPROXYMITY;

	while (attempt < MAX_PROX_ATTEMPT) {
		attempt++;

		answer = NULL;
		answer_size = 0;
		do {
			status = process_proximity(MS_DRM_CONTEXT(device), answer, answer_size, &message, &message_size);
			if (status != RM_PENDING){
				/* Error and we need to restart or success and
				 * we need to stop*/
				break;
			} else {
				/* Send a new proximity message */
				/* TODO this breaks the select loop, add it inside the ILibChain */
				/* Try to send and receive, polling for 50 ms before retrying, and during 120 s */
				status = ms_udp_send_receive(ms_socket, message, message_size, &answer, &answer_size, 50, 120);
				if (RMFAILED(status))
					break;
			}
		} while (RMSUCCEEDED(status));

		if (RMSUCCEEDED(status))
			/* Success, proximity is over */
			break;
	}
	
	if (!(attempt < MAX_PROX_ATTEMPT)){
		RMDBGLOG((ENABLE,"Too many attempt, proximity test failed\n"));
		goto error;
	}
	
	RMDBGLOG((ENABLE,"Proximity detection OK\n"));

	close_ms_udp_socket(ms_socket);

	/* We are not waiting anything anymore */
	MS_UPNP(device)->flags &= ~MS_WAITPROXYMITY;
	/* Promixity OK */
	MS_UPNP(device)->flags |= MS_PROXYMITYOK;
	return RM_OK;

error:
	if (ms_socket)
		close_ms_udp_socket(ms_socket);

	/* We are not waiting anything anymore */
	MS_UPNP(device)->flags &= ~MS_WAITPROXYMITY;
	return RM_ERROR;
}
	






	
